"""
wuyang.py

Functions associated with wuyang inversion
"""

import numpy as np
from opt_einsum import contract
from scipy.optimize import minimize

class WuYang():
    """
    Performs Optimization as in: 10.1063/1.1535422 - Qin Wu + Weitao Yang

    Attributes:
    -----------
    lambda_rgl: {None, float}. If float, lambda-regularization is added with lambda=lambda_rgl.
    """

    regul_norm = None  # Regularization norm: ||v||^2
    lambda_reg = None  # Regularization constant

    def wuyang(self, opt_max_iter, reg=None, tol=1e-7, gtol=1e-3,
               opt_method='trust-krylov', opt=None):
        """
        Calls scipy minimizer to minimize lagrangian. 
        """
        self.lambda_reg = reg
        if opt is None:
            opt = {"disp"    : False}
        opt['maxiter'] = opt_max_iter
        opt['gtol'] = gtol
        # Initialization for D and C
        self._diagonalize_with_potential_pbs(self.v_pbs)

        if opt_method.lower() == 'bfgs' or opt_method.lower() == 'l-bfgs-b':
            opt_results = minimize( fun = self.lagrangian_wy,
                                    x0  = self.v_pbs,
                                    jac = self.gradient_wy,
                                    method  = opt_method,
                                    tol     = tol,
                                    options = opt
                                    )

        else:
            opt_results = minimize( fun = self.lagrangian_wy,
                                    x0  = self.v_pbs,
                                    jac = self.gradient_wy,
                                    hess = self.hessian_wy,
                                    method = opt_method,
                                    tol    = tol,
                                    options = opt
                                    )

        if opt_results.success == False:
            self.v_pbs = opt_results.x
            self.opt_info = opt_results
            raise ValueError("Optimization was unsucessful (|grad|=%.2e) within %i iterations, "
                             "try a different initial guess. %s"% (np.linalg.norm(opt_results.jac), opt_results.nit, opt_results.message)
                             )
        else:
            print("Optimization Successful within %i iterations! "
                  "|grad|=%.2e" % (opt_results.nit, np.linalg.norm(opt_results.jac)))
            self.v_pbs = opt_results.x
            self.opt_info = opt_results

    def _diagonalize_with_potential_pbs(self, v):
        """
        Diagonalize Fock matrix with additional external potential
        """
        self.v_pbs = np.copy(v)
        vks_a = contract("ijk,k->ij", self.S3, v[:self.npbs]) + self.va
        fock_a = self.V + self.T + vks_a 
        self.Ca, self.Coca, self.Da, self.eigvecs_a = self.diagonalize( fock_a, self.nalpha )

        if self.ref == 1:
            self.Cb, self.Cocb, self.Db, self.eigvecs_b = self.Ca.copy(), self.Coca.copy(), self.Da.copy(), self.eigvecs_a.copy()
            self.Fock =  fock_a
        else:
            vks_b = contract("ijk,k->ij", self.S3, v[self.npbs:]) + self.vb
            fock_b = self.V + self.T + vks_b        
            self.Cb, self.Cocb, self.Db, self.eigvecs_b = self.diagonalize( fock_b, self.nbeta )
            self.Fock =  (fock_a, fock_b)

    def lagrangian_wy(self, v):
        """
        Lagrangian to be minimized wrt external potential
        Equation (5) of main reference
        """
        # If v is not updated, will not re-calculate.
        if not np.allclose(v, self.v_pbs):
            self._diagonalize_with_potential_pbs(v)

        self.grad_a = contract('ij,ijt->t', (self.Da - self.Dt[0]), self.S3)
        self.grad_b = contract('ij,ijt->t', (self.Db - self.Dt[1]), self.S3)

        kinetic     =   np.sum(self.T * (self.Da))
        potential   =   np.sum((self.V + self.va) * (self.Da - self.Dt[0]))
        optimizing  =   np.sum(v[:self.npbs] * self.grad_a)

        if self.ref == 1:
            L = 2 * (kinetic + potential + optimizing)

        else:
            kinetic    +=   np.sum(self.T * (self.Db))
            potential  +=   np.sum((self.V + self.vb) * (self.Db - self.Dt[1]))
            optimizing +=   np.sum(v[self.npbs:] * self.grad_b)
            L = kinetic + potential + optimizing

        # Add lambda-regularization
        if self.lambda_reg is not None:
            T = self.T_pbs
            if self.ref == 1:
                norm = 2 * (v[:self.npbs] @ T @ v[:self.npbs])
            else:
                norm = (v[self.npbs:] @ T @ v[self.npbs:]) + (v[:self.npbs] @ T @ v[:self.npbs])

            L -= norm * self.lambda_reg
            self.regul_norm = norm

        # if print_flag:
        #    print(f"Kinetic: {kinetic:6.4f} | Potential: {np.abs(potential):6.4e} | From Optimization: {np.abs(optimizing):6.4e}")

        return - L

    def gradient_wy(self, v):
        """
        Calculates gradient wrt target density
        Equation (11) of main reference
        """
        if not np.allclose(v, self.v_pbs):
            self._diagonalize_with_potential_pbs(v)
        self.grad_a = contract('ij,ijt->t', (self.Da - self.Dt[0]), self.S3)
        self.grad_b = contract('ij,ijt->t', (self.Db - self.Dt[1]), self.S3)

        if self.ref == 1:
            self.grad   = self.grad_a
        else:
            self.grad   = np.concatenate(( self.grad_a, self.grad_b ))

        if self.lambda_reg is not None:
            T = self.T_pbs
            if self.ref == 1:
                rgl_vector = 4 * self.lambda_reg*np.dot(T, v[:self.npbs])
                self.grad -= rgl_vector
            else:
                self.grad[:self.npbs] -= 2 * self.lambda_reg*np.dot(T, v[:self.npbs])
                self.grad[self.npbs:] -= 2 * self.lambda_reg*np.dot(T, v[self.npbs:])

        return -self.grad

    def hessian_wy(self, v):
        """
        Calculates gradient wrt target density
        Equation (13) of main reference
        """

        if not np.allclose(v, self.v_pbs):
            self._diagonalize_with_potential_pbs(v)

        na, nb = self.nalpha, self.nbeta

        eigs_diff_a = self.eigvecs_a[:na, None] - self.eigvecs_a[None, na:]
        C3a = contract('mi,va,mvt->iat', self.Ca[:,:na], self.Ca[:,na:], self.S3)
        Ha = 2 * contract('iau,iat,ia->ut', C3a, C3a, eigs_diff_a**-1)

        if self. ref == 1:
            if self.lambda_reg is not None:
                Ha -= 4 * self.T_pbs * self.lambda_reg
            Hs = Ha

        else:

            eigs_diff_b = self.eigvecs_b[:nb, None] - self.eigvecs_b[None, nb:]
            C3b = contract('mi,va,mvt->iat', self.Cb[:,:nb], self.Cb[:,nb:], self.S3)
            Hb = 2 * contract('iau,iat,ia->ut', C3b, C3b, eigs_diff_b**-1)
            if self.lambda_reg is not None:
                Ha -= 2 * self.T_pbs * self.lambda_reg
                Hb -= 2 * self.T_pbs * self.lambda_reg
            Hs = np.block(
                            [[Ha,                               np.zeros((self.npbs, self.npbs))],
                            [np.zeros((self.npbs, self.npbs)), Hb                              ]]
                        )

        return - Hs

    def find_regularization_constant_wy(self, opt_max_iter, opt_method="trust-krylov", gtol=1e-3,
                                     tol=None, opt=None, lambda_list=None):
        """
        Finding regularization constant lambda.

        Note: it is recommend to set a specific convergence criteria by opt or tol,
                in order to control the same convergence
                for different lambda value.

        After the calculation is done, one can plot the returns to select a good lambda.

        Parameters:
        -----------
        opt_max_iter: int
                    maximum iteration

        opt_method: string default: "trust-krylov"
            opt_methods available in scipy.optimize.minimize

        tol: float
            Tolerance for termination. See scipy.optimize.minimize for details.
        gtol: float
             gtol for scipy.optimize.minimize: the gradient norm for
             convergence
        opt: dictionary, optional
            if given:
                scipy.optimize.minimize(method=opt_method, options=opt).
            Notice that opt has lower priorities than opt_max_iter and gtol.

        lambda_list: np.ndarray, optional
            A array of lambda to search; otherwise, it will be 10 ** np.linspace(-1, -7, 7).

        Returns:
        --------
        lambda_list: np.ndarray
            A array of lambda searched.

        P_list: np.ndarray
            The value defined by [Bulat, Heaton-Burgess, Cohen, Yang 2007] eqn (21).
            Corresponding to lambda in lambda_list.

        Ts_list: np.ndarray
            The Ts value for each lambda.


        """

        Ts_list = []
        L_list = []
        v_norm_list = []


        if lambda_list is None:
            lambda_list = 10 ** np.linspace(-3, -9, 7)

        if opt is None:
            opt = {"disp"    : False}
        opt['maxiter'] = opt_max_iter
        opt['gtol'] = gtol

        self.lambda_reg = None
        # Initial calculation with no regularization
        # Initialization for D and C
        self._diagonalize_with_potential_pbs(self.v_pbs)

        if opt_method.lower() == 'bfgs' or opt_method.lower() == 'l-bfgs-b':
            initial_result = minimize(fun=self.lagrangian_wy,
                                   x0=self.v_pbs,
                                   jac=self.gradient_wy,
                                   method=opt_method,
                                   tol=tol,
                                   options=opt
                                   )
        else:
            initial_result = minimize(fun=self.lagrangian_wy,
                                   x0=self.v_pbs,
                                   jac=self.gradient_wy,
                                   hess=self.hessian_wy,
                                   method=opt_method,
                                   tol=tol,
                                   options=opt
                                   )
        if initial_result.success == False:
            raise ValueError("Optimization was unsucessful (|grad|=%.2e) within %i iterations, "
                             "try a different intitial guess"% (np.linalg.norm(initial_result.jac), initial_result.nit)
                             + initial_result.message)
        else:
            L0 = -initial_result.fun
            initial_v0 = initial_result.x  # This is used as the initial guess for with regularization calculation.

        for reg in lambda_list:
            self.lambda_reg = reg

            if opt_method.lower() == 'bfgs' or opt_method.lower() == 'l-bfgs-b':
                opt_results = minimize(fun=self.lagrangian_wy,
                                       x0=initial_v0,
                                       jac=self.gradient_wy,
                                       method=opt_method,
                                       tol=tol,
                                       options=opt
                                       )

            else:
                opt_results = minimize(fun=self.lagrangian_wy,
                                       x0=initial_v0,
                                       jac=self.gradient_wy,
                                       hess=self.hessian_wy,
                                       method=opt_method,
                                       tol=tol,
                                       options=opt
                                       )


            Ts_list.append(np.sum(self.T * (self.Da + self.Db)))
            v_norm_list.append(self.regul_norm)
            L_list.append(-opt_results.fun + self.lambda_reg * self.regul_norm)

        P_list = lambda_list * np.array(v_norm_list) / (L0 - np.array(L_list))

        return lambda_list, P_list, np.array(Ts_list)



"""
inverter.py
Density-to-potential inversion module

Handles the primary functions
"""

import numpy as np
from dataclasses import dataclass
from opt_einsum import contract

try:
    import psi4
    psi4.core.be_quiet()
    psi4.core.clean()
except:
    pass    

from .methods.wuyang import WuYang
from .methods.zmp import ZMP
from .methods.mrks import MRKS
from .methods.oucarter import OC
from .methods.pdeco import PDECO
from .methods.direct import Direct
from .grid.grider import Grider

@dataclass
class data_bucket:
    """
    Data class for storing grid attributes. 
    """
    pass


class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
    """
    Attributes:
    ----------
    wfn : psi4.core.{RHF, UHF, RKS, UKS, Wavefunction, CCWavefuncion...}
        Psi4 wavefunction object
    mol : psi4.core.Molecule
        Psi4 molecule object
    basis : psi4.core.BasisSet
        Psi4 basis set object
    basis_str : str
        Basis set
    nbf : int
        Number of basis functions for main calculation
    nalpha : int
        Number of alpha electrons
    nbeta : int
        Number of beta electrons
    ref : {1,2}
        Reference calculation
        1 -> Restricted
        2 -> Unrestricted
    Dt : List
        List of psi4.core.Matrix for target density matrices (on AO).
    ct : List
        List of psi4.core.Matrix for input occupied orbitals. This might not be correct for post-HartreeFock methods.
    pbs_str: string
        name of Potential basis set
    pbs : psi4.core.BasisSet
        Potential basis set.
    npbs : int
        the length of pbs
    v_pbs : np.ndarray shape (npbs, ) for ref==1 and (2*npbs, ) for ref==2.
        potential vector on the Potential Baiss Set.
        If the potential is not represented on the basis set, this should
        remain 0. It will be initialized to a 0 array. One can set this
        value for initial guesses before Wu-Yang method (WY) or PDE-Constrained
        Optimization method (PDE-CO). For example, if PDE-CO is ran after
        a WY calculation, the initial for PDE-CO will be the result of WY
        if v_pbs is not zeroed.
    S2  : np.ndarray
        The ao overlap matrix (i.e. S matrix)
    S3  : np.ndarray
        The three ao overlap matrix (ao, ao, pbs)
    S4  : np.ndarray
        The four ao overlap matrix, the size should be (ao, ao, ao, ao)
    jk  : psi4.core.JK
        Psi4 jk object. Built if wfn has no jk, otherwise use wfn.jk

    T   : np.ndarray
        kinetic matrix on ao
    V   : np.ndarray
        external potential matrix on ao
    T_pbs: np.ndarray
        kinetic matrix on pbs. Useful for regularization.

    guide_potential_components: list of string
        guide potential components name
    va, vb: np.ndarray of shape (nbasis, nbasis)
        guide potential Fock matrix.
    Methods:
    --------

    """
    def __init__(self, wfn, pbs="same", debug=False):
        """
        Handles Inversion
        
        Parameters
        ----------
        wfn : psi4.core.{RHF, UHF, RKS, UKS, Wavefunction, CCWavefuncion...}
            Psi4 wavefunction object
        pbs: str. default="same". If same, then the potential basis set (pbs)
                 is the same as orbital basis set (i.e. ao). Notice that
                 pbs is not needed for some methods
        """
        self.wfn       = wfn
        self.pbs_str   = pbs
        self.mol       = wfn.molecule()
        self.basis     = wfn.basisset()
        self.basis_str = wfn.basisset().name()
        self.nbf       = wfn.basisset().nbf()
        self.nalpha    = wfn.nalpha()
        self.nbeta     = wfn.nbeta()
        self.ref       = 1 if psi4.core.get_global_option("REFERENCE") == "RHF" or \
                              psi4.core.get_global_option("REFERENCE") == "RKS" else 2
        self.jk        = wfn.jk() if hasattr(wfn, "jk") == True else self.generate_jk()
        self.Dt        = (np.array(wfn.Da_subset("AO")), np.array(wfn.Db_subset("AO")))
        self.ct        = (np.array(wfn.Ca_subset("AO", "OCC")), np.array(wfn.Cb_subset("AO", "OCC")))
        self.pbs       = self.basis if pbs == "same" \
                                    else psi4.core.BasisSet.build( self.mol, key='BASIS', target=self.pbs_str)
        self.npbs      = self.pbs.nbf()
        self.v_pbs     = np.zeros( (self.npbs) ) if self.ref == 1 \
                                                 else np.zeros( 2 * self.npbs )
        self.generate_mints_matrices()
        self.grid = data_bucket
        self.cubic_grid = data_bucket
        
        self.J0 = None
        self.S4 = None  # Entry to save the 4 overlap matrix.

    #------------->  Basics:

    def generate_mints_matrices(self):
        """
        Generates matrices that are methods of a mints object
        """

        mints = psi4.core.MintsHelper( self.basis )

        #Overlap Matrices
        self.S2 = np.array(mints.ao_overlap())
        A = mints.ao_overlap()
        A.power( -0.5, 1e-16 )
        self.A = np.array(A)
        self.S3 = np.squeeze(mints.ao_3coverlap(self.basis,self.basis,self.pbs))

        #Core Matrices
        self.T = np.array(mints.ao_kinetic()).copy()
        self.V = np.array(mints.ao_potential()).copy()
        self.T_pbs = np.array(mints.ao_kinetic(self.pbs, self.pbs)).copy()

    def generate_jk(self, gen_K=False):
        """
        Creates jk object for generation of Coulomb and Exchange matrices
        1.0e9 B -> 1.0 GB
        """
        jk = psi4.core.JK.build(self.basis)
        memory = int(jk.memory_estimate() * 1.1)
        jk.set_memory(int(memory)) 
        jk.set_do_K(gen_K)
        jk.initialize()

        return jk

    def form_jk(self, Cocc_a, Cocc_b):
        """
        Generates Coulomb and Exchange matrices from occupied orbitals
        """
        # if  self.jk.memory_estimate() > psi4.get_memory() * 0.8:
        #     raise ValueError("Requested JK will take too more memory than default. \n \
        #                       Increase it with psi4.set_memory(int( Many More Bytes )))!")
        Cocc_a = psi4.core.Matrix.from_array(Cocc_a)
        Cocc_b = psi4.core.Matrix.from_array(Cocc_b)
        self.jk.C_left_add(Cocc_a)
        self.jk.C_left_add(Cocc_b) 
        self.jk.compute()
        self.jk.C_clear()

        J = (np.array(self.jk.J()[0]), np.array(self.jk.J()[1]))
        K = []

        return J, K

    def diagonalize(self, matrix, ndocc):
        """
        Diagonalizes Fock Matrix

        Parameters
        ----------
        marrix: np.ndarray
            Matrix to be diagonalized
        ndocc: int
            Number of occupied orbitals

        Returns
        -------
        C: np.ndarray
            Orbital Matrix
        Cocc: np.ndarray
            Occupied Orbital Matrix
        D: np.ndarray
            Density Matrix
        eigves: np.ndarray
            Eigenvalues
        """

        A = self.A
        Fp = A.dot(matrix).dot(A)
        eigvecs, Cp = np.linalg.eigh(Fp)
        C = A.dot(Cp)
        Cocc = C[:, :ndocc]
        D = contract('pi,qi->pq', Cocc, Cocc)
        return C, Cocc, D, eigvecs

    #------------->  Inversion:

    def invert(self, method,
                     guide_potential_components = ["fermi_amaldi"],
                     opt_max_iter = 50,
                     **keywords):
        """
        Handler to all available inversion methods

        Parameters
        ----------

        method: str
            Method used to invert density. 
            Can be chosen from {wuyang, zmp, mrks, oc}. 
            See documentation below for each method. 
        guide_potential_components: list, opt
            Components added as to guide inversion. 
            Can be chosen from {"fermi_amandi", "svwn"}
            Default: ["fermi_amaldi"]
        opt_max_iter: int, opt
            Maximum number of iterations inside the chosen inversion.
            Default: 50

        direct
        ------
        Direct inversion of a set of Kohn-Sham equations. 
        $$v_{xc}(r) = \frac{1}{n(r)} \sum_i^N [\phi_i^{*} (r) \nabla^2 \phi_i(r) + \varepsilon_i | \phi_i(r)|^2] $$
            Parameters:
            -----------
                grid: np.ndarray, opt
                    Grid where result will be expressed in.
                    If not provided, dft grid will be used instead. 
                
        wuyang
        ------
        the Wu-Yang method:
        The Journal of chemical physics 118.6 (2003): 2498-2509.
            Parameters:
            ----------
                opt_max_iter: int
                    maximum iteration
                opt_method: string, opt
                    Method for scipy optimizer
                    Currently only used by wuyang and pdeco method.
                    Defaul: 'trust-krylov'
                    https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
                reg : float, opt
                    Regularization constant for Wuyant Inversion.
                    Default: None -> No regularization is added.
                    Becomes attribute of inverter -> inverter.lambda_reg
                tol: float
                    tol for scipy.optimize.minimize
                gtol: float
                    gtol for scipy.optimize.minimize: the gradient norm for
                    convergence
                opt: dict
                    options for scipy.optimize.minimize
                                Notice that opt has lower priorities than opt_max_iter and gtol.
            return:
                the result are stored in self.v_pbs

            
        zmp
        ---
        The Zhao-Morrison-Parr Method:
        Phys. Rev. A 50, 2138 
            Parameters:
            ----------
                lambda_list: list
                    List of Lamda parameters used as a coefficient for Hartree 
                    difference in SCF cycle.
                zmp_mixing: float, optional
                    mixing \in [0,1]. How much of the new potential is added in.
                    For example, zmp_mixing = 0 means the traditional ZMP, i.e. all the potentials from previous
                    smaller lambda are ignored.
                    Zmp_mixing = 1 means that all the potentials of previous lambdas are accumulated, the larger lambda
                    potential are meant to fix the wrong/inaccurate region of the potential of the sum of the previous
                    potentials instead of providing an entire new potentials.
                    default: 1
                opt_max_iter: float
                    Maximum number of iterations for scf cycle
                opt_tol: float
                    Convergence criteria set for Density Difference and DIIS error. 
                return:
                    The result will be stored in self.proto_density_a and self.proto_density_b
                    For zmp_mixing==1, restricted (ref==1):
                        self.proto_density_a = \sum_i lambda_i * (Da_i - Dt[0]) - 1/N * (Dt[0])
                        self.proto_density_b = \sum_i lambda_i * (Db_i - Dt[1]) - 1/N * (Dt[1]);
                    unrestricted (ref==1):
                        self.proto_density_a = \sum_i lambda_i * (Da_i - Dt[0]) - 1/N * (Dt[0] + Dt[1])
                        self.proto_density_b = \sum_i lambda_i * (Db_i - Dt[1]) - 1/N * (Dt[0] + Dt[1]);
                    For restricted (ref==1):
                        vxc = \int dr' \frac{self.proto_density_a + self.proto_density_b}{|r-r'|}
                            = 2 * \int dr' \frac{self.proto_density_a}{|r-r'|};
                    for unrestricted (ref==2):
                        vxc_up = \int dr' \frac{self.proto_density_a}{|r-r'|}
                        vxc_down = \int dr' \frac{self.proto_density_b}{|r-r'|}.
                    To get potential on grid, one needs to do
                        vxc = self.on_grid_esp(Da=self.proto_density_a, Db=self.proto_density_b, grid=grid) for restricted;
                        vxc_up = self.on_grid_esp(Da=self.proto_density_a, Db=np.zeros_like(self.proto_density_a),
                                  grid=grid) for unrestricted;


        mRKS
        ----
        the modified Ryabinkin-Kohut-Staroverov method:
        Phys. Rev. Lett. 115, 083001 
        J. Chem. Phys. 146, 084103p
            Parameters:
            -----------
                maxiter: int
                    same as opt_max_iter
                vxc_grid: np.ndarray of shape (3, num_grid_points), opt
                    When this is given, the final result will be represented
                v_tol: float, opt
                    convergence criteria for vxc Fock matrices.
                    default: 1e-4
                D_tol: float, opt
                    convergence criteria for density matrices.
                    default: 1e-7
                eig_tol: float, opt
                    convergence criteria for occupied eigenvalue spectrum.
                    default: 1e-4
                frac_old: float, opt
                    Linear mixing parameter for current vxc and old vxc.
                    If 0, no old vxc is mixed in.
                    Should be in [0,1)
                    default: 0.5.
                init: string or psi4.core.Wavefunction, opt
                    Initial guess method.
                    default: "SCAN"
                    1) If None, input wfn info will be used as initial guess.
                    2) If "continue" is given, then it will not initialize
                    but use the densities and orbitals stored. Meaningly,
                    one can run a quick WY calculation as the initial
                    guess. This can also be used to user speficified
                    initial guess by setting Da, Coca, eigvec_a.
                    3) If it's not continue, it would be expecting a
                    method name string that works for psi4. A separate psi4 calculation
                    would be performed.
                sing: tuple of float of length 4, opt.
                    Singularity parameter for _vxc_hole_quadrature()
                    default: (1e-5, 1e-4, 1e-5, 1e-4)
                    [0]: atol, [1]: atol1 for dft_spherical grid calculation.
                    [2]: atol, [3]: atol1 for vxc_grid calculation.
            return:
                The result will be stored in self.grid.vxc

        oc
        --
        Ou-Carter method
        J. Chem. Theory Comput. 2018, 14, 5680−5689
            Parameters:
            -----------
                maxiter: int
                    same as opt_max_iter
                vxc_grid: np.ndarray of shape (3, num_grid_points)
                    The final result will be represented on this grid
                    default: 1e-4
                D_tol: float, opt
                    convergence criteria for density matrices.
                    default: 1e-7
                eig_tol: float, opt
                    convergence criteria for occupied eigenvalue spectrum.
                    default: 1e-4
                frac_old: float, opt
                    Linear mixing parameter for current vxc and old vxc.
                    If 0, no old vxc is mixed in.
                    Should be in [0,1)
                    default: 0.5.
                init: string, opt
                    Initial guess method.
                    default: "SCAN"
                    1) If None, input wfn info will be used as initial guess.
                    2) If "continue" is given, then it will not initialize
                    but use the densities and orbitals stored. Meaningly,
                    one can run a quick WY calculation as the initial
                    guess. This can also be used to user speficified
                    initial guess by setting Da, Coca, eigvec_a.
                    3) If it's not continue, it would be expecting a
                    method name string that works for psi4. A separate psi4 calculation
                    would be performed.
                    wuyang

        pdeco
        -----
        the PDE-Constrained Optimization method:
        Int J Quantum Chem. 2018;118:e25425;
        Nat Commun 10, 4497 (2019).
            Parameters:
            ----------
                opt_max_iter: int
                    maximum iteration
                opt_method: string, opt
                    Method for scipy optimizer
                    Currently only used by wuyang and pdeco method.
                    Defaul: 'L-BFGS-B'
                    Options: ['L-BFGS-B', 'BFGS']
                    https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
                reg : float, opt
                    Regularization constant for Wuyant Inversion.
                    Default: None -> No regularization is added.
                    Becomes attribute of inverter -> inverter.lambda_reg
                gtol: float
                    gtol for scipy.optimize.minimize: the gradient norm for
                    convergence
                opt: dict
                    options for scipy.optimize.minimize
                    Notice that opt has lower priorities than opt_max_iter and gtol.
            return:
                the result are stored in self.v_pbs
        """

        #Generate Guide Potential
        if method.lower()=='mrks':
            if guide_potential_components[0] != 'hartree' or len(guide_potential_components) != 1:
                print("The guide potential is changed to v_hartree.")
            self.generate_components(["hartree"])
        elif method.lower() != 'direct':
            self.generate_components(guide_potential_components)

        #Invert
        if method.lower() == "direct":
            return self.direct_inversion(**keywords)
        elif method.lower() == "wuyang":
            self.wuyang(opt_max_iter, **keywords)
        elif method.lower() == "zmp":
            self.zmp(opt_max_iter, **keywords)
        elif method.lower() == "mrks":
            return self.mRKS(opt_max_iter, **keywords)
        elif method.lower() == 'oc':
            return self.oucarter(opt_max_iter, **keywords)
        elif method.lower() == 'pdeco':
            return self.pdeco(opt_max_iter, **keywords)
        else:
            raise ValueError(f"Inversion method not available. Methods available: {['wuyang', 'zmp', 'mrks', 'oc', 'pdeco']}")

    def generate_components(self, guide_potential_components):
        """
        Generates exact potential components to be added to
        the Hamiltonian to aide in the inversion procedure. 

        Parameters:
        -----------
        guide_potential_components: list
            Components added as to guide inversion. 
            Can be chosen from {"fermi_amandi", "svwn"}
        """

        self.va = np.zeros_like(self.T)
        self.vb = np.zeros_like(self.T)

        N = self.nalpha + self.nbeta

        if self.J0 is None:
            if type(self.wfn) == psi4.core.CCWavefunction:
                nbf = self.nbf
                C_NO = psi4.core.Matrix(nbf, nbf)
                eigs_NO = psi4.core.Vector(nbf)
                self.wfn.Da().diagonalize(C_NO, eigs_NO, psi4.core.DiagonalizeOrder.Descending)
                occu = np.sqrt(eigs_NO.np)
                New_Orb_a = occu * C_NO.np
                assert np.allclose(New_Orb_a @ New_Orb_a.T, self.Dt[0])
                if self.ref == 1:
                    New_Orb_b = New_Orb_a
                else:
                    self.wfn.Db().diagonalize(C_NO, eigs_NO, psi4.core.DiagonalizeOrder.Descending)
                    occu = np.sqrt(eigs_NO.np)
                    New_Orb_b = occu * C_NO.np
                self.J0 = self.form_jk(New_Orb_a, New_Orb_b)[0]
            else:
                self.J0, _ = self.form_jk(self.ct[0], self.ct[1])

        if "fermi_amaldi" in guide_potential_components:
            v_fa = (1-1/N) * (self.J0[0] + self.J0[1])

            self.va += v_fa
            self.vb += v_fa
        elif "hartree" in guide_potential_components:
            v_hartree = (self.J0[0] + self.J0[1])

            self.va += v_hartree
            self.vb += v_hartree
        else:
            raise ValueError("Hartee nor FA was included."
                             "Convergence will likely not be achieved")

        if "svwn" in guide_potential_components:
            if "svwn" in guide_potential_components:
                _, wfn_0 = psi4.energy( "svwn"+"/"+self.basis_str, molecule=self.mol , return_wfn = True)

            if self.ref == 1:
                ntarget = psi4.core.Matrix.from_array( [ self.Dt[0] + self.Dt[1] ] )
                wfn_0.V_potential().set_D( [ntarget] )
                vxc_a = psi4.core.Matrix( self.nbf, self.nbf )
                wfn_0.V_potential().compute_V([vxc_a])
                self.va += vxc_a.np
                self.vb += vxc_a.np
            elif self.ref == 2:
                na_target = psi4.core.Matrix.from_array( self.Dt[0] )
                nb_target = psi4.core.Matrix.from_array( self.Dt[1] )
                wfn_0.V_potential().set_D( [na_target, nb_target] )
                vxc_a = psi4.core.Matrix( self.nbf, self.nbf )
                vxc_b = psi4.core.Matrix( self.nbf, self.nbf )
                wfn_0.V_potential().compute_V([vxc_a, vxc_b])
                self.va += vxc_a.np
                self.vb += vxc_b.np
        self.guide_potential_components = guide_potential_components



import n2v
import psi4
import pytest

psi4.set_options({"save_jk" : True})
psi4.set_memory(int(2.50e9))


@pytest.fixture
def wfn():
    Ne = psi4.geometry( 
    """ 
    0 1
    Ne
    noreorient
    nocom
    units bohr
    symmetry c1
    """ )
    psi4.set_options({"reference" : "rhf"}) 
    wfn = psi4.properties("svwn/cc-pvdz",  return_wfn=True, molecule=Ne, property=['dipole'])[1]

    return wfn

def test_sucess_inversion(wfn):
    ine = n2v.Inverter(wfn, pbs="cc-pvtz")
    ine.invert("WuYang", guide_potential_components=["fermi_amaldi"])

    assert ine.opt_info.success == True




import n2v
import numpy as np
import psi4
import pytest

psi4.set_options({"save_jk" : True})
psi4.set_memory(int(2.50e9))

psi4.set_options({
                  'DFT_SPHERICAL_POINTS': 50, 
                  'DFT_RADIAL_POINTS': 50, 
                  "opdm": True,
                  "tpdm": True,
                 })

@pytest.fixture
def wfn():
    Ne = psi4.geometry( 
    """ 
    0 1
    Ne 0.0 0.0 0.0
    noreorient
    nocom
    units bohr
    symmetry c1
    """ )
    psi4.set_options({"reference" : "rhf"}) 
    wfn = psi4.properties("detci/cc-pcvdz", return_wfn=True, molecule=Ne, properties=["dipole"])[1]


    return wfn

def test_sucess_inversion(wfn):

    x = np.linspace(-5,10,1501)
    y = [0]
    z = [0]


    ine = n2v.Inverter(wfn)
    grid, shape = ine.generate_grids(x,y,z)
    ine.invert("mRKS", opt_max_iter=30, frac_old=0.8, init="scan")

    correct_eigs_a = np.array([-30.57378503,-1.61491607,-0.75653413,-0.75653413,-0.75653413])

    assert np.isclose(ine.eigvecs_a[:5].all(), correct_eigs_a.all())



"""
grider.py
"""

import numpy as np
import psi4

from opt_einsum import contract

class Psi4Grider():
    def __init__(self, mol, basis, ref):
        
        self.mol = mol
        self.ref = ref
        self.basis = basis
        self.basis_str = basis.name()
        self.nbf   = self.basis.nbf()

        wfn_base = psi4.core.Wavefunction.build(self.mol, self.basis_str)
        self.wfn = psi4.proc.scf_wavefunction_factory('svwn', wfn_base, "UKS")
        self.wfn.initialize()

        self.nalpha = self.wfn.nalpha()
        self.nbeta  = self.wfn.nbeta()

        # Clean Vpot
        restricted = True if ref == 1 else False
        reference  = "RV" if ref == 1 else "UV"
        functional = psi4.driver.dft.build_superfunctional("SVWN", restricted=restricted)[0]
        self.Vpot = psi4.core.VBase.build(self.basis, functional, reference)
        self.Vpot.initialize()

        self.npoints = self.Vpot.grid().npoints()   

    def compute_LDA(self):
        """
        Computres LDA Calculation on Molecule
        """
        wfn = psi4.energy("SVWN/" + self.basis_str, molecule=self.mol, return_wfn=True)[1]
        Da  = wfn.Da().np
        Db  = wfn.Db().np
        Ca  = wfn.Ca().np
        Cb  = wfn.Cb().np
        epsilon_a = wfn.epsilon_a().np
        epsilon_b = wfn.epsilon_b().np

        return Da, Db, Ca, Cb, epsilon_a, epsilon_b

    def generate_grid(self, x, y, z):
        """
        Genrates Mesh from 3 separate linear spaces and flatten,
        needed for cubic grid.

        Parameters
        ----------
        grid: tuple of three np.ndarray
            (x, y, z)

        Returns
        -------
        grid: np.ndarray
            shape (3, len(x)*len(y)*len(z)).
        """
        # x,y,z, = grid
        shape = (len(x), len(y), len(z))
        X,Y,Z = np.meshgrid(x, y, z, indexing='ij')
        X = X.reshape((X.shape[0] * X.shape[1] * X.shape[2], 1))
        Y = Y.reshape((Y.shape[0] * Y.shape[1] * Y.shape[2], 1))
        Z = Z.reshape((Z.shape[0] * Z.shape[1] * Z.shape[2], 1))
        grid = np.concatenate((X,Y,Z), axis=1).T

        return grid, shape

    def grid_to_blocks(self, grid, basis=None):
        """
        Generate list of blocks to allocate given grid

        Parameters
        ----------
        grid: np.ndarray
            Grid to be distributed into blocks
            Size: (3, npoints) for homogeneous grid
                  (4, npoints) for inhomogenous grid to account for weights
        basis: psi4.core.BasisSet; optional
            The basis set. If not given, it will use target wfn.basisset().

        Returns
        -------
        blocks: list    
            List with psi4.core.BlockOPoints
        npoints: int
            Total number of points (for one dimension)
        points: psi4.core.{RKS, UKS}
            Points function to set matrices.
        """
        assert (grid.shape[0] == 3) or (grid.shape[0] == 4), """Grid does not have the correct dimensions. \n
                                                              Array must be of size (3, npoints) or (4, npoints)"""
        if_w = grid.shape[0] == 4

        if basis is None:
            basis = self.basis

        epsilon    = psi4.core.get_global_option("CUBIC_BASIS_TOLERANCE")
        extens     = psi4.core.BasisExtents(basis, epsilon)
        max_points = psi4.core.get_global_option("DFT_BLOCK_MAX_POINTS")        
        npoints    = grid.shape[1]
        nblocks = int(np.floor(npoints/max_points))
        blocks = []

        max_functions = 0
        #Run through full blocks
        idx = 0
        for nb in range(nblocks):
            x = psi4.core.Vector.from_array(grid[0][idx : idx + max_points])
            y = psi4.core.Vector.from_array(grid[1][idx : idx + max_points])
            z = psi4.core.Vector.from_array(grid[2][idx : idx + max_points])
            if if_w:
                w = psi4.core.Vector.from_array(grid[3][idx : idx + max_points])
            else:
                w = psi4.core.Vector.from_array(np.zeros(max_points))  # When w is not necessary and not given

            blocks.append(psi4.core.BlockOPoints(x, y, z, w, extens))
            idx += max_points
            max_functions = max_functions if max_functions > len(blocks[-1].functions_local_to_global()) \
                                          else len(blocks[-1].functions_local_to_global())

        #Run through remaining points
        if idx < npoints:
            x = psi4.core.Vector.from_array(grid[0][idx:])
            y = psi4.core.Vector.from_array(grid[1][idx:])
            z = psi4.core.Vector.from_array(grid[2][idx:])
            if if_w:
                w = psi4.core.Vector.from_array(grid[3][idx:])
            else:
                w = psi4.core.Vector.from_array(np.zeros_like(grid[2][idx:]))  # When w is not necessary and not given
            blocks.append(psi4.core.BlockOPoints(x, y, z, w, extens))

            max_functions = max_functions if max_functions > len(blocks[-1].functions_local_to_global()) \
                                          else len(blocks[-1].functions_local_to_global())

        zero_matrix = psi4.core.Matrix(basis.nbf(), basis.nbf())
        if self.ref == 1:
            point_func = psi4.core.RKSFunctions(basis, max_points, max_functions)
            point_func.set_pointers(zero_matrix)
        else:
            point_func = psi4.core.UKSFunctions(basis, max_points, max_functions)
            point_func.set_pointers(zero_matrix, zero_matrix)

        return blocks, npoints, point_func

    def ao(self, coeff, grid=None, basis=None):
        """
        Generates a quantity on the grid given its ao representation.
        *This is the most general function for basis to grid transformation.

        Parameters
        ----------
        coeff: np.ndarray
            Vector/Matrix of quantity on ao basis. Shape: {(num_ao_basis, ), (num_ao_basis, num_ao_basis)}
        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
            grid where density will be computed.
        basis: psi4.core.BasisSet, optional
            The basis set. If not given it will use target wfn.basisset().
        Vpot: psi4.core.VBase
            Vpotential object with info about grid. 
            Provides DFT spherical grid. Only comes to play if no grid is given. 

        Returns
        -------
        coeff_r: np.ndarray Shape: (npoints, )
            Quantity expressed by the coefficient on the given grid 


        """


        if grid is not None:
            if type(grid) is np.ndarray:
                if grid.shape[0] != 3 and grid.shape[0] != 4:
                    raise ValueError("The shape of grid should be (3, npoints) "
                                     "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
                blocks, npoints, points_function = self.grid_to_blocks(grid, basis=basis)
            else:
                blocks, npoints, points_function = grid
        elif grid is None:
            Vpot = self.Vpot
            nblocks = Vpot.nblocks()
            blocks = [Vpot.get_block(i) for i in range(nblocks)]
            npoints = Vpot.grid().npoints()
            points_function = Vpot.properties()[0]
        else:
            raise ValueError("A grid or a V_potential (DFT grid) must be given.")

        coeff_r = np.zeros((npoints))

        offset = 0
        for i_block in blocks:
            points_function.compute_points(i_block)
            b_points = i_block.npoints()
            offset += b_points
            lpos = np.array(i_block.functions_local_to_global())
            if len(lpos)==0:
                continue
            phi = np.array(points_function.basis_values()["PHI"])[:b_points, :lpos.shape[0]]

            if coeff.ndim == 1:
                l_mat = coeff[(lpos[:])]
                coeff_r[offset - b_points : offset] = contract('pm,m->p', phi, l_mat)
            elif coeff.ndim == 2:
                l_mat = coeff[(lpos[:, None], lpos)]
                coeff_r[offset - b_points : offset] = contract('pm,mn,pn->p', phi, l_mat, phi)

        return coeff_r

    def density(self, Da, Db=None, grid=None):
        """
        Generates Density given grid

        Parameters
        ----------
        Da, Db: np.ndarray
            Alpha, Beta densities. Shape: (num_ao_basis, num_ao_basis)
        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
            grid where density will be computed.
        vpot: psi4.core.VBase
            Vpotential object with info about grid.
            Provides DFT spherical grid. Only comes to play if no grid is given. 

        Returns
        -------
        density: np.ndarray Shape: (ref, npoints)
            Density on the given grid. 
        """

        Da = psi4.core.Matrix.from_array(Da)
        if Db is not None:
            Db = psi4.core.Matrix.from_array(Db)

        if grid is not None:
            if type(grid) is np.ndarray:
                if grid.shape[0] != 3 and grid.shape[0] != 4:
                    raise ValueError("The shape of grid should be (3, npoints) "
                                     "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
                blocks, npoints, points_function = self.grid_to_blocks(grid)
            else:
                blocks, npoints, points_function = grid
        elif grid is None:
            nblocks = self.Vpot.nblocks()
            blocks = [self.Vpot.get_block(i) for i in range(nblocks)]
            npoints = self.Vpot.grid().npoints()
            points_function = self.Vpot.properties()[0]
        else:
            raise ValueError("A grid or a V_potential (DFT grid) must be given.")

        if Db is None:
            points_function.set_pointers(Da)
            rho_a = points_function.point_values()["RHO_A"]
            density   = np.zeros((npoints))
        else:
            points_function.set_pointers(Da, Db)
            rho_a = points_function.point_values()["RHO_A"]
            rho_b = points_function.point_values()["RHO_B"]
            density   = np.zeros((npoints, self.ref))

        offset = 0
        for i_block in blocks:
            points_function.compute_points(i_block)
            b_points = i_block.npoints()
            offset += b_points

            if self.ref == 1:
                density[offset - b_points : offset] = rho_a.np[ :b_points]
            else:
                density[offset - b_points : offset, 0] = rho_a.np[ :b_points]
                density[offset - b_points : offset, 1] = rho_b.np[ :b_points]

        return density

        pass

    def esp(self, Da=None, Db=None, vpot=None, grid=None, compute_hartree=True):
        """
        Generates EXTERNAL/ESP/HARTREE and Fermi Amaldi Potential on given grid
        Parameters
        ----------
        Da,Db: np.ndarray, opt, shape (nbf, nbf)
            The electron density in the denominator of Hartee potential. If None, the original density matrix
            will be used.
        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
            grid where density will be computed.
        vpot: psi4.core.VBase
            Vpotential object with info about grid.
            Provides DFT spherical grid. Only comes to play if no grid is given. 
        
        Returns
        -------
        vext, hartree, esp, v_fa: np.ndarray
            External, Hartree, ESP, and Fermi Amaldi potential on the given grid
            Shape: (npoints, )
        """
        wfn = self.wfn

        if Da is not None or Db is not None:
            Da_temp = np.copy(self.wfn.Da().np)
            Db_temp = np.copy(self.wfn.Db().np)
            if Da is not None:
                wfn.Da().np[:] = Da
            if Db is not None:
                wfn.Db().np[:] = Db

        nthreads = psi4.get_num_threads()
        psi4.set_num_threads(1)
        
        if grid is not None:
            if type(grid) is np.ndarray:
                blocks, npoints, points_function = self.grid_to_blocks(grid)
            else:
                blocks, npoints, points_function = grid
        elif vpot is not None:
            nblocks = vpot.nblocks()
            blocks = [vpot.get_block(i) for i in range(nblocks)]
            npoints = vpot.grid().npoints()
        elif vpot is not None and grid is not None:
            raise ValueError("Only one option can be given")

        #Initialize Arrays
        vext = np.zeros(npoints)
        esp  = np.zeros(npoints)

        #Get Atomic Information
        mol_dict = self.mol.to_schema(dtype='psi4')
        natoms = len(mol_dict["elem"])
        indx = [i for i in range(natoms) if self.mol.charge(i) != 0.0]
        natoms = len(indx)
        #Atomic numbers and Atomic positions
        zs = [mol_dict["elez"][i] for i in indx]
        rs = [self.mol.geometry().np[i] for i in indx]

        if compute_hartree:
            esp_wfn = psi4.core.ESPPropCalc(wfn)

        #Loop Through blocks
        offset = 0
        with np.errstate(divide='ignore'):
            for i_block in blocks:
                b_points = i_block.npoints()
                offset += b_points
                x = i_block.x().np
                y = i_block.y().np
                z = i_block.z().np

                #EXTERNAL
                for atom in range(natoms):
                    r =  np.sqrt((x-rs[atom][0])**2 + (y-rs[atom][1])**2 + (z-rs[atom][2])**2)
                    vext_temp = - 1.0 * zs[atom] / r
                    vext_temp[np.isinf(vext_temp)] = 0.0
                    vext[offset - b_points : offset] += vext_temp

                if compute_hartree:
                    #ESP
                    xyz = np.concatenate((x[:,None],y[:,None],z[:,None]), axis=1)
                    grid_block = psi4.core.Matrix.from_array(xyz)
                    esp[offset - b_points : offset] = esp_wfn.compute_esp_over_grid_in_memory(grid_block).np

        #Hartree
        if compute_hartree:
            hartree = - 1.0 * (vext + esp)

        if Da is not None:
            self.wfn.Da().np[:] = Da_temp
        if Db is not None:
            self.wfn.Db().np[:] = Db_temp
        psi4.set_num_threads(nthreads)

        if compute_hartree:
            return vext, hartree, esp
        else:
            return vext

    def vxc(self, func_id=1, Da=None, Db=None, grid=None):
        """
        Generates Vxc given grid
        Parameters
        ----------
        Da, Db: np.ndarray
            Alpha, Beta densities. Shape: (num_ao_basis, num_ao_basis)
        func_id: int
            Functional ID associated with Density Functional Approximationl.
            Full list of functionals: https://www.tddft.org/programs/libxc/functionals/
        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
            grid where density will be computed.
        Vpot: psi4.core.VBase
            Vpotential object with info about grid.
            Provides DFT spherical grid. Only comes to play if no grid is given. 
        Returns
        -------
        VXC: np.ndarray
            Exchange correlation potential on the given grid
            Shape: (npoints, )
        """

        if func_id != 1:
            raise ValueError("Only LDA fucntionals are supported on the grid")

        if Da is None:
            Da = self.Dt[0]
        if Db is None:
            Db = self.Dt[0]

        if grid is not None:
            if type(grid) is np.ndarray:
                blocks, npoints, points_function = self.grid_to_blocks(grid)
            else:
                blocks, npoints, points_function = grid
            density = self.density(Da=Da, Db=Db, grid=grid)
        elif grid is None and Vpot is not None:
            nblocks = self.Vpot.nblocks()
            blocks = [Vpot.get_block(i) for i in range(nblocks)]
            npoints = Vpot.grid().npoints()
            density = self.density(Da=Da, Db=Db, Vpot=self.Vpot)
        else:
            raise ValueError("A grid or a V_potential (DFT grid) must be given.")

        vxc = np.zeros((npoints, self.ref))
        ingredients = {}
        offset = 0
        for i_block in blocks:
            b_points = i_block.npoints()
            offset += b_points
            if self.ref == 1:
                ingredients["rho"] = density[offset - b_points : offset]
            else:
                ingredients["rho"] = density[offset - b_points : offset, :]

            if self.ref == 1:
                functional = Functional(1, 1)
            else:
                functional = Functional(1, 2) 
            xc_dictionary = functional.compute(ingredients)
            vxc[offset - b_points : offset, :] = xc_dictionary['vrho']

        return np.squeeze(vxc)

    def average_local_orbital_energy(self, D, C, eig, grid_info=None):
        """
        (4)(6) in mRKS.
        """

        # Nalpha = self.molecule.nalpha
        # Nbeta = self.molecule.nbeta

        if grid_info is None:
            e_bar = np.zeros(self.Vpot.grid().npoints())
            nblocks = self.Vpot.nblocks()

            points_func = self.Vpot.properties()[0]
            points_func.set_deriv(0)
            blocks = None
        else:
            blocks, npoints, points_func = grid_info
            e_bar = np.zeros(npoints)
            nblocks = len(blocks)
            points_func.set_deriv(0)

        if Db is None:
            points_func.set_pointers(Da)
        else:
            points_function.set_pointers(Da, Db)

        # For unrestricted
        iw = 0
        for l_block in range(nblocks):
            # Obtain general grid information
            if blocks is None:
                l_grid = self.Vpot.get_block(l_block)
            else:
                l_grid = blocks[l_block]
            l_npoints = l_grid.npoints()

            points_func.compute_points(l_grid)
            l_lpos = np.array(l_grid.functions_local_to_global())
            if len(l_lpos) == 0:
                iw += l_npoints
                continue
            l_phi = np.array(points_func.basis_values()["PHI"])[:l_npoints, :l_lpos.shape[0]]
            lD = D[(l_lpos[:, None], l_lpos)]
            lC = C[l_lpos, :]
            rho = contract('pm,mn,pn->p', l_phi, lD, l_phi)
            e_bar[iw:iw + l_npoints] = contract("pm,mi,ni,i,pn->p", l_phi, lC, lC, eig, l_phi) / rho

            iw += l_npoints
        assert iw == e_bar.shape[0], "Somehow the whole space is not fully integrated."
        return e_bar

    def optimized_external_potential(self, grid=None, average_alpha_beta=False):
        """
        $
        v^{~}{ext}(r) = \epsilon^{-LDA}(r)
        - \frac{\tau^{LDA}{L}}{n^{LDA}(r)}
        - v_{H}^{LDA}(r) - v_{xc}^{LDA}(r)
        $
        (22) in [1].
        """

        Nalpha = self.nalpha
        Nbeta = self.nbeta

        # LDA calculation
        Da_LDA, Db_DLA, Ca_LDA, Cb_LDA, epsilon_a_LDA, epsilon_b_LDA = self.eng.compute_LDA()
        vxc_LDA = self.on_grid_vxc(Da=Da_LDA, Db=Db_LDA, grid=grid)

        if self.ref != 1:
            assert vxc_LDA.shape[-1] == 2
            vxc_LDA_beta = vxc_LDA[:,1]
            vxc_LDA = vxc_LDA[:, 0]
            vxc_LDA_DFT_beta = vxc_LDA_DFT[:, 1]
            vxc_LDA_DFT = vxc_LDA_DFT[:, 0]

        # _average_local_orbital_energy() taken from mrks.py.
        #---> Compute average loca orbital energy
        e_bar = self.average_local_orbital_energy(Da_LDA, Ca_LDA[:, :Nalpha], epsilon_a_LDA[:Nalpha], grid=grid)
        tauLmP_rho = self._get_l_kinetic_energy_density_directly(Da_LDA, Ca_LDA[:,:Nalpha], grid=grid_info)
        tauP_rho = self._pauli_kinetic_energy_density(Da_LDA, Ca_LDA[:,:Nalpha], grid=grid)

        tauL_rho_DFT = tauLmP_rho_DFT + tauP_rho_DFT
        tauL_rho = tauLmP_rho + tauP_rho

        vext_opt_no_H_DFT = e_bar_DFT - tauL_rho_DFT - vxc_LDA_DFT
        vext_opt_no_H = e_bar - tauL_rho - vxc_LDA

        J = self.form_jk(Ca_LDA[:,:Nalpha],  Cb_LDA[:,:Nbeta])[0]
        vext_opt_no_H_DFT_Fock = self.dft_grid_to_fock(vext_opt_no_H_DFT, self.Vpot)
        vext_opt_DFT_Fock = vext_opt_no_H_DFT_Fock - J[0] - J[1]

        # # Does vext_opt need a shift?
        # Fock_LDA = self.T + vext_opt_DFT_Fock + J[0] + J[1] + self.dft_grid_to_fock(vxc_LDA_DFT, self.Vpot)
        # eigvecs_a = self.diagonalize(Fock_LDA, self.nalpha)[-1]
        # shift = eigvecs_a[Nalpha-1] - epsilon_a_LDA[Nalpha-1]
        # vext_opt_DFT_Fock -= shift * self.S2
        # print("LDA shift:", shift, eigvecs_a[Nalpha-1], epsilon_a_LDA[Nalpha-1])

        vH = self.on_grid_esp(grid=grid_info, wfn=wfn_LDA)[1]
        vext_opt = vext_opt_no_H - vH
        # vext_opt -= shift

        if self.ref != 1:
            e_bar_DFT_beta = self._average_local_orbital_energy(Db_LDA, Cb_LDA[:,:Nbeta], epsilon_b_LDA[:Nbeta])
            e_bar_beta = self._average_local_orbital_energy(Db_LDA, Cb_LDA[:, :Nbeta], epsilon_b_LDA[:Nbeta], grid_info=grid_info)


            tauLmP_rho_DFT_beta = self._get_l_kinetic_energy_density_directly(Db_LDA, Cb_LDA[:,:Nbeta])
            tauLmP_rho_beta = self._get_l_kinetic_energy_density_directly(Db_LDA, Cb_LDA[:,:Nalpha], grid_info=grid_info)

            tauP_rho_DFT_beta = self._pauli_kinetic_energy_density(Db_LDA, Cb_LDA[:,:Nbeta])
            tauP_rho_beta = self._pauli_kinetic_energy_density(Db_LDA, Cb_LDA[:,:Nbeta], grid_info=grid_info)

            tauL_rho_DFT_beta = tauLmP_rho_DFT_beta + tauP_rho_DFT_beta
            tauL_rho_beta = tauLmP_rho_beta + tauP_rho_beta

            vext_opt_no_H_DFT_beta = e_bar_DFT_beta - tauL_rho_DFT_beta - vxc_LDA_DFT_beta
            vext_opt_no_H_beta = e_bar_beta - tauL_rho_beta - vxc_LDA_beta

            vext_opt_no_H_DFT_Fock_beta = self.dft_grid_to_fock(vext_opt_no_H_DFT_beta, self.Vpot)
            vext_opt_DFT_Fock_beta = vext_opt_no_H_DFT_Fock_beta - J[0] - J[1]

            vext_opt_beta = vext_opt_no_H_beta - vH

            # vext_opt_DFT_Fock = (vext_opt_DFT_Fock + vext_opt_DFT_Fock_beta) * 0.5
            # vext_opt = (vext_opt + vext_opt_beta) * 0.5

            return (vext_opt_DFT_Fock, vext_opt_DFT_Fock_beta), (vext_opt, vext_opt_beta)
        return vext_opt_DFT_Fock, vext_opt

    def get_orbitals(self, basis, molecule):

        pass



import n2v
import psi4
import numpy as np
import pytest

psi4.set_options({"save_jk" : True})
psi4.set_memory(int(2.50e9))


@pytest.fixture
def Ne():
    Ne = psi4.geometry( 
    """ 
    0 1
    Ne
    noreorient
    nocom
    units bohr
    symmetry c1
    """ )
    psi4.set_options({"reference" : "rhf"}) 
    wfn = psi4.properties("svwn/cc-pvdz",  return_wfn=True, molecule=Ne, property=['dipole'])[1]

    ine = n2v.Inverter(wfn, pbs="cc-pvtz")
    ine.invert("WuYang", guide_potential_components=["fermi_amaldi"])

    x = np.linspace(-5,5,5)
    y = [0]
    z = [0]
    grid, shape = ine.generate_grids(x,y,z)

    return (ine, wfn, grid)

def test_density( Ne ):

    ine = Ne[0]
    wfn = Ne[1]
    grid = Ne[2]

    da = ine.on_grid_density(Da=wfn.Da().np, Db=wfn.Db(), grid=grid)
    target = np.array([1.09245597e-09, 3.20317361e-03, 5.92592390e+02, 3.20317361e-03,1.09245597e-09])

    assert np.isclose(da, target).all()

def test_coeff( Ne ):

    ine = Ne[0]
    wfn = Ne[1]
    grid = Ne[2]

    da = ine.on_grid_ao(coeff=wfn.Da().np, grid=grid)
    target = np.array([5.46227987e-10, 1.60158681e-03, 2.96296195e+02, 1.60158681e-03, 5.46227987e-10])

    assert np.isclose(da, target).all()

def test_coeff_single( Ne ):

    ine = Ne[0]
    wfn = Ne[1]
    grid = Ne[2]

    rest = ine.on_grid_ao(coeff=ine.v_pbs, grid=grid)
    target = np.array([-2.92277751e-07, -2.69322822e-03, -5.15373654e+00, -2.69322822e-03, -2.92277751e-07])

    assert np.isclose(rest, target).all()

def test_esp( Ne ):

    ine = Ne[0]
    wfn = Ne[1]
    grid = Ne[2]

    vext, hartree, v_fa, esp = ine.on_grid_esp(Da=wfn.Da().np, Db=wfn.Db().np, grid=grid)
    t_ext = np.array([ -2., -4.,  0., -4., -2. ])
    t_har = np.array([  2.        ,  3.9973889 , 31.06080946,  3.9973889 ,  2.         ])
    t_fa = np.array([ 1.8       ,  3.59765001, 27.95472852,  3.59765001,  1.8 ])
    t_esp = np.array([ 1.96521466e-10,  2.61109825e-03, -3.10608095e+01,  2.61109825e-03,1.96521466e-10 ])

    assert np.isclose( vext, t_ext ).all()
    assert np.isclose( hartree, t_har ).all()
    assert np.isclose( v_fa, t_fa ).all()
    assert np.isclose( esp, t_esp ).all()

def test_vxc( Ne ):

    ine = Ne[0]
    wfn = Ne[1]
    grid = Ne[2]

    vxc = ine.on_grid_vxc(Da=wfn.Da().np, Db=wfn.Db().np, grid=grid)
    target = np.array([-1.01420351e-03, -1.45161223e-01, -8.27133894e+00, -1.45161223e-01,-1.01420351e-03])

    assert np.isclose(vxc, target).all()


def test_orbitals( Ne ):

    ine = Ne[0]
    wfn = Ne[1]
    grid = Ne[2]

    orbitals = ine.on_grid_orbitals(Ca=wfn.Ca().np, Cb=wfn.Cb().np, grid=grid)
    target_0 = np.array([-1.39426057e-08, -1.27790509e-04,  1.67530855e+01, -1.27790509e-04, -1.39426057e-08])
    target_1 = np.array([-1.25081898e-06, -1.15539283e-02,  3.95352024e+00, -1.15539283e-02, -1.25081898e-06])

    print(orbitals[0])
    print(orbitals[1])

    assert np.isclose(orbitals[0], target_0, atol=1e-3).all()
    assert np.isclose(orbitals[1], target_1, atol=1e-3).all()

def test_lap_phi( Ne ):

    ine = Ne[0]
    wfn = Ne[1]
    grid = Ne[2]

    lap = ine.on_grid_lap_phi(Ca=wfn.Ca().np, Cb=wfn.Cb().np, grid=grid)
    target_0 = np.array([-2.89807670e-07, -3.40206526e-04, -1.26411748e+05, -3.40206526e-04, -2.89807670e-07])
    target_1 = np.array([-2.59992244e-05, -3.57517732e-02, -3.00612855e+04, -3.57517732e-02,-2.59992244e-05])

    print(lap[0])
    print(lap[1])

    assert np.isclose(lap[0], target_0).all()
    assert np.isclose(lap[1], target_1).all()

def test_grad_phi( Ne ):

    ine = Ne[0]
    wfn = Ne[1]
    grid = Ne[2]

    grad = ine.on_grid_grad_phi(Ca=wfn.Ca().np, Cb=wfn.Cb().np, grid=grid)
    target_2 = np.array([ -4.41609823e-05, -1.84240102e-02,  1.47088832e+01, -1.84240102e-02, -4.41609823e-05 ])
    target_10 = np.array([ -8.20775970e-20, -2.26659995e-04, -5.71751099e-15,  2.26659995e-04, -9.48517230e-20 ])

    # assert np.isclose(grad[2], target_2).all()
    # assert np.isclose(grad[10], target_10).all()

def test_int_ao( Ne ):

    ine = Ne[0]
    wfn = Ne[1]
    grid = Ne[2]

    density = ine.on_grid_density(Da=wfn.Da().np, Db=wfn.Db().np, Vpot=wfn.V_potential())
    int_density = ine.dft_grid_to_fock( value=density, Vpot=wfn.V_potential() )

    pass





import n2v
import numpy as np
import psi4
import pytest

psi4.set_options({"save_jk" : True})
psi4.set_memory(int(2.50e9))


@pytest.fixture
def wfn():
    Ne = psi4.geometry( 
    """ 
    0 1
    Ne
    noreorient
    nocom
    units bohr
    symmetry c1
    """ )
    psi4.set_options({"reference" : "rhf"}) 
    wfn = psi4.properties("ccsd/aug-cc-pvtz",  return_wfn=True, molecule=Ne, property=['dipole'])[1]

    return wfn

def test_sucess_inversion(wfn):
    ine = n2v.Inverter(wfn, pbs="aug-cc-pvqz")
    ine.invert("PDECO", opt_max_iter=200, guide_potential_components=["fermi_amaldi"], gtol=1e-6)

    correct_eigs_a = np.array([-30.824053,-1.57094445,-0.71606582,-0.71586213,-0.71584401])

    assert np.isclose(ine.eigvecs_a[:5].all(), correct_eigs_a.all())



"""
zmp.py

Functions associated with zmp inversion
"""


import psi4
psi4.core.be_quiet()
import numpy as np
from functools import reduce


eps = np.finfo(float).eps

class ZMP():
    def zmp(self, 
            opt_max_iter=100, 
            opt_tol= psi4.core.get_option("SCF", "D_CONVERGENCE"), 
            lambda_list=[70],
            zmp_mixing = 1,
            print_scf = False, 
            ):

        """
        Performs ZMP optimization according to: 
        
        1) 'From electron densities to Kohn-Sham kinetic energies, orbital energies,
        exchange-correlation potentials, and exchange-correlation energies' by
        Zhao + Morrison + Parr. 
        https://doi.org/10.1103/PhysRevA.50.2138

        Additional DIIS algorithms obtained from:
        2) 'Psi4NumPy: An interactive quantum chemistry programming environment 
        for reference implementations and rapid development.' by 
        Daniel G.A. Smith and others. 
        https://doi.org/10.1021/acs.jctc.8b00286

        Functionals that drive the SCF procedure are obtained from:
        https://doi.org/10.1002/qua.26400

        Parameters:
        -----------
        lambda_list: list
            List of Lamda parameters used as a coefficient for Hartree
            difference in SCF cycle.
        zmp_mixing: float, optional
            mixing \in [0,1]. How much of the new potential is added in.
            For example, zmp_mixing = 0 means the traditional ZMP, i.e. all the potentials from previous
            smaller lambda are ignored.
            Zmp_mixing = 1 means that all the potentials of previous lambdas are accumulated, the larger lambda
            potential are meant to fix the wrong/inaccurate region of the potential of the sum of the previous
            potentials instead of providing an entire new potentials.
            default: 1
        opt_max_iter: float
            Maximum number of iterations for scf cycle
        opt_tol: float
            Convergence criteria set for Density Difference and DIIS error.
        return:
            The result will be stored in self.proto_density_a and self.proto_density_b
            For zmp_mixing==1, restricted (ref==1):
                self.proto_density_a = \sum_i lambda_i * (Da_i - Dt[0]) - 1/N * (Dt[0])
                self.proto_density_b = \sum_i lambda_i * (Db_i - Dt[1]) - 1/N * (Dt[1]);
            unrestricted (ref==1):
                self.proto_density_a = \sum_i lambda_i * (Da_i - Dt[0]) - 1/N * (Dt[0] + Dt[1])
                self.proto_density_b = \sum_i lambda_i * (Db_i - Dt[1]) - 1/N * (Dt[0] + Dt[1]);
            For restricted (ref==1):
                vxc = \int dr' \frac{self.proto_density_a + self.proto_density_b}{|r-r'|}
                    = 2 * \int dr' \frac{self.proto_density_a}{|r-r'|};
            for unrestricted (ref==2):
                vxc_up = \int dr' \frac{self.proto_density_a}{|r-r'|}
                vxc_down = \int dr' \frac{self.proto_density_b}{|r-r'|}.
            To get potential on grid, one needs to do
                vxc = self.on_grid_esp(Da=self.proto_density_a, Db=self.proto_density_b, grid=grid) for restricted;
                vxc_up = self.on_grid_esp(Da=self.proto_density_a, Db=np.zeros_like(self.proto_density_a),
                          grid=grid) for unrestricted;
        """
        self.diis_space = 100
        self.mixing = zmp_mixing

        print("\nRunning ZMP:")
        self.zmp_scf(lambda_list, 'hartree', opt_max_iter, print_scf, D_conv=opt_tol)

    def zmp_scf(self, 
            lambda_list,
            zmp_functional,
            maxiter, 
            print_scf,
            D_conv):
        """
        Performs scf cycle
        Parameters:
            zmp_functional: options the penalty term.
            But others are not currently working except for Hartree penalty (original ZMP).
        ----------
        """

        #Checks if there is dft grid available. 
        if hasattr(self.wfn, 'V_potential()') == False:
            ref = "RV" if self.ref == 1 else "UV"
            functional = psi4.driver.dft.build_superfunctional("SVWN", restricted=True if self.ref==1 else False)[0]
            Vpot = psi4.core.VBase.build(self.basis, functional, ref)
            Vpot.initialize()
            self.Vpot = Vpot
        else:
            self.Vpot = self.wfn.V_potential()

        # Obtain target density on dft_grid 
        D0 = self.on_grid_density(grid=None, Da=self.Dt[0], Db=self.Dt[1],  Vpot=self.Vpot)

        #Calculate kernels
        if zmp_functional.lower() != 'hartree':
            if self.ref == 1:
                Da0, Db0 = D0[:,0], D0[:,0]
            else:
                Da0, Db0 = D0[:,0], D0[:,1]

            if zmp_functional == 'log':
                Sa0 = np.log(Da0) + 1  
                Sb0 = np.log(Db0) + 1
            elif zmp_functional == 'exp':
                Sa0 = Da0 ** 0.05
                Sb0 = Db0 ** 0.05
            self.Sa0 = self.dft_grid_to_fock(Sa0, Vpot=self.Vpot)
            if self.ref == 2:
                self.Sb0 = self.dft_grid_to_fock(Sb0, Vpot=self.Vpot)
            else: 
                self.Sb0 = self.Sa0.copy()

        vc_previous_a = np.zeros((self.nbf, self.nbf))
        vc_previous_b = np.zeros((self.nbf, self.nbf))
        self.Da = self.Dt[0]
        self.Db = self.Dt[1]
        Da = self.Dt[0]
        Db = self.Dt[1]
        Cocca = self.ct[0]
        Coccb = self.ct[1]

        grid_diff_old = 1/np.finfo(float).eps
        self.proto_density_a =  np.zeros_like(Da)
        self.proto_density_b =  np.zeros_like(Db)

#------------->  Iterating over lambdas:
        for lam_i in lambda_list:
            self.shift = 0.1 * lam_i
            D_old = self.Dt[0]

            # Trial & Residual Vector Lists
            state_vectors_a, state_vectors_b = [], []
            error_vectors_a, error_vectors_b = [], []

            for SCF_ITER in range(1,maxiter):

#------------->  Generate Fock Matrix:
                vc = self.generate_s_functional(lam_i,
                                                zmp_functional,
                                                Cocca, Coccb, 
                                                Da, Db)

                #Equation 10 of Reference (1). Level shift. 
                Fa = self.T + self.V + self.va + vc[0] + vc_previous_a
                Fa += (self.S2 - reduce(np.dot, (self.S2, Da, self.S2))) * self.shift

                if self.ref == 2:
                    Fb = self.T + self.V + self.vb + vc[1] + vc_previous_b
                    Fb += (self.S2 - reduce(np.dot, (self.S2, Db, self.S2))) * self.shift
                

    #------------->  DIIS:
                if SCF_ITER > 1:
                    #Construct the AO gradient
                    # r = (A(FDS - SDF)A)_mu_nu
                    grad_a = self.A.dot(Fa.dot(Da).dot(self.S2) - self.S2.dot(Da).dot(Fa)).dot(self.A)
                    grad_a[np.abs(grad_a) < eps] = 0.0

                    if SCF_ITER < self.diis_space:
                        state_vectors_a.append(Fa.copy())
                        error_vectors_a.append(grad_a.copy())
                    else:
                        state_vectors_a.append(Fa.copy())
                        error_vectors_a.append(grad_a.copy())
                        del state_vectors_a[0]
                        del error_vectors_a[0]

                    #Build inner product of error vectors
                    Bdim = len(state_vectors_a)
                    Ba = np.empty((Bdim + 1, Bdim + 1))
                    Ba[-1, :] = -1
                    Ba[:, -1] = -1
                    Ba[-1, -1] = 0
                    Bb = Ba.copy()

                    for i in range(len(state_vectors_a)):
                        for j in range(len(state_vectors_a)):
                            Ba[i,j] = np.einsum('ij,ij->', error_vectors_a[i], error_vectors_a[j])

                    #Build almost zeros matrix to generate inverse. 
                    RHS = np.zeros(Bdim + 1)
                    RHS[-1] = -1

                    #Find coefficient matrix:
                    Ca = np.linalg.solve(Ba, RHS.copy())
                    Ca[np.abs(Ca) < eps] = 0.0 

                    #Generate new fock Matrix:
                    Fa = np.zeros_like(Fa)
                    for i in range(Ca.shape[0] - 1):
                        Fa += Ca[i] * state_vectors_a[i]

                    diis_error_a = np.max(np.abs(error_vectors_a[-1]))
                    if self.ref ==  1:
                        Fb = Fa.copy()
                        diis_error = 2 * diis_error_a

                    else:
                        grad_b = self.A.dot(Fb.dot(Db).dot(self.S2) - self.S2.dot(Db).dot(Fb)).dot(self.A)
                        grad_b[np.abs(grad_b) < eps] = 0.0
                        
                        if SCF_ITER < self.diis_space:
                            state_vectors_b.append(Fb.copy())
                            error_vectors_b.append(grad_b.copy())
                        else:
                            state_vectors_b.append(Fb.copy())
                            error_vectors_b.append(grad_b.copy())
                            del state_vectors_b[0]
                            del error_vectors_b[0]

                        for i in range(len(state_vectors_b)):
                            for j in range(len(state_vectors_b)):
                                Bb[i,j] = np.einsum('ij,ij->', error_vectors_b[i], error_vectors_b[j])

                        diis_error_b = np.max(np.abs(error_vectors_b[-1]))
                        diis_error = diis_error_a + diis_error_b

                        Cb = np.linalg.solve(Bb, RHS.copy())
                        Cb[np.abs(Cb) < eps] = 0.0 

                        Fb = np.zeros_like(Fb)
                        for i in range(Cb.shape[0] - 1):
                            Fb += Cb[i] * state_vectors_b[i]

                else:
                    diis_error = 1.0
                
    #------------->  Diagonalization | Check convergence:

                Ca, Cocca, Da, eigs_a = self.diagonalize(Fa, self.nalpha)
                if self.ref == 2:
                    Cb, Coccb, Db, eigs_b = self.diagonalize(Fb, self.nbeta)
                else: 
                    Cb, Coccb, Db, eigs_b = Ca.copy(), Cocca.copy(), Da.copy(), eigs_a.copy()

                ddm = D_old - Da
                D_old = Da
                derror = np.max(np.abs(ddm))

                #Uncomment to debug internal SCF
                if print_scf is True:
                    if np.mod(SCF_ITER,5) == 0.0:
                        print(f"Iteration: {SCF_ITER:3d} | Self Convergence Error: {derror:10.5e} | DIIS Error: {diis_error:10.5e}")
                
                #DIIS error may improve as fast as the D_conv. Relax the criteria an order of magnitude. 
                if abs(derror) < D_conv and abs(diis_error) < D_conv*10: 
                    break
                if SCF_ITER == maxiter - 1:
                    raise ValueError("ZMP Error: Maximum Number of SCF cycles reached. Try different settings.")

            density_current = self.on_grid_density(grid=None, Da=Da, Db=Db, Vpot=self.Vpot)
            grid_diff = np.max(np.abs(D0 - density_current))
            if np.abs(grid_diff_old) < np.abs(grid_diff):
                # This is a greedy algorithm: if the density error stopped improving for this lambda, we will stop here.
                print(f"\nZMP halted at lambda={lam_i}. Density Error Stops Updating: old: {grid_diff_old}, current: {grid_diff}.")
                break

            grid_diff_old = grid_diff
            print(f"SCF Converged for lambda:{int(lam_i):5d}. Max density difference: {grid_diff}")

            self.proto_density_a += lam_i * (Da - self.Dt[0]) * self.mixing
            if self.ref == 2:
                self.proto_density_b += lam_i * (Db - self.Dt[1]) * self.mixing
            else:
                self.proto_density_b = self.proto_density_a.copy()

            vc_previous_a += vc[0] * self.mixing
            if self.ref == 2:
                vc_previous_b += vc[1] * self.mixing

            # this is the lambda that is already proven to be improving the density, i.e. the corresponding
            # potential has updated to proto_density
            successful_lam = lam_i
            # The proto_density corresponds to successful_lam
            successful_proto_density = [(Da - self.Dt[0]), (Db - self.Dt[1])]
# -------------> END Iterating over lambdas:

        self.proto_density_a += successful_lam * successful_proto_density[0] * (1 - self.mixing)
        if self.guide_potential_components[0].lower() == "fermi_amaldi":
            # for ref==1, vxc = \int dr (proto_density_a + proto_density_b)/|r-r'| - 1/N*vH
            if self.ref == 1:
                self.proto_density_a -= (1 / (self.nalpha + self.nbeta)) * (self.Dt[0])
            # for ref==1, vxc = \int dr (proto_density_a)/|r-r'| - 1/N*vH
            else:
                self.proto_density_a -= (1 / (self.nalpha + self.nbeta)) * (self.Dt[0] + self.Dt[1])

        self.Da = Da
        self.Ca = Ca
        self.Coca = Cocca
        self.eigvecs_a = eigs_a

        if self.ref == 2:
            self.proto_density_b += successful_lam * successful_proto_density[1] * (1 - self.mixing)
            if self.guide_potential_components[0].lower() == "fermi_amaldi":
                # for ref==1, vxc = \int dr (proto_density_a + proto_density_b)/|r-r'| - 1/N*vH
                if self.ref == 1:
                    self.proto_density_b -= (1 / (self.nalpha + self.nbeta)) * (self.Dt[1])
                # for ref==1, vxc = \int dr (proto_density_a)/|r-r'| - 1/N*vH
                else:
                    self.proto_density_b -= (1 / (self.nalpha + self.nbeta)) * (self.Dt[0] + self.Dt[1])
            self.Db = Db
            self.Cb = Cb
            self.Cocb = Coccb
            self.eigvecs_b = eigs_b

        else:
            self.proto_density_b = self.proto_density_a.copy()
            self.Db = self.Da.copy()
            self.Cb = self.Ca.copy()
            self.Cocb = self.Coca.copy()
            self.eigvecs_b = self.eigvecs_a.copy()
        print(successful_lam)


    def generate_s_functional(self, lam, zmp_functional, Cocca, Coccb, Da, Db):
        """
        Generates S_n Functional as described in:
        https://doi.org/10.1002/qua.26400
        """

        if zmp_functional.lower() == 'hartree':
            J, _ = self.form_jk(Cocca, Coccb)

            #Equation 7 of Reference (1)
            if self.ref == 1:
                vc_a = 2 * lam * ( J[0] - self.J0[0] ) 
                vc = [vc_a]
            else:
                vc_a = lam * ( J[0] - self.J0[0] ) 
                vc_b = lam * ( J[1] - self.J0[1] )
                vc = [vc_a, vc_b]            
            return vc

        else:

            D = self.on_grid_density(Vpot=self.Vpot, Da=Da, Db=Db)

            if self.ref == 1:
                Da, Db = D[:,0], D[:,0]
            else:
                Da, Db = D[:,0], D[:,1]
                
            #Calculate kernels
            #Equations 37, 38, 39, 40 of Reference (3)
            if zmp_functional.lower() == 'log':
                Sa = np.log(Da) + 1
                Sb = np.log(Db) + 1 
            if zmp_functional.lower() == 'exp':
                Sa = Da ** 0.05
                Sb = Db ** 0.05

            #Generate AO matrix from functions. 
            Sa = self.dft_grid_to_fock(Sa, Vpot=self.Vpot)
            Sb = self.dft_grid_to_fock(Sb, Vpot=self.Vpot)
            vc = (Sa + Sb) - (self.Sa0 + self.Sb0)

        return vc
        



"""
grider.py

Generates grid for plotting 
"""

import numpy as np
import warnings
from opt_einsum import contract
import psi4
psi4.core.be_quiet()

try:
    from pylibxc import LibXCFunctional as Functional
except:
    pass

# from .cubeprop import Cubeprop
from .basis_set_artifact_correction import basis_set_correction, invert_kohn_sham_equations

class Grider():

    def grid_to_blocks(self, grid, basis=None):
        """
        Generate list of blocks to allocate given grid

        Parameters
        ----------
        grid: np.ndarray
            Grid to be distributed into blocks
            Size: (3, npoints) for homogeneous grid
                  (4, npoints) for inhomogenous grid to account for weights
        basis: psi4.core.BasisSet; optional
            The basis set. If not given, it will use target wfn.basisset().

        Returns
        -------
        blocks: list    
            List with psi4.core.BlockOPoints
        npoints: int
            Total number of points (for one dimension)
        points: psi4.core.{RKS, UKS}
            Points function to set matrices.
        """
        assert (grid.shape[0] == 3) or (grid.shape[0] == 4), """Grid does not have the correct dimensions. \n
                                                              Array must be of size (3, npoints) or (4, npoints)"""
        if_w = grid.shape[0] == 4

        if basis is None:
            basis = self.basis

        epsilon    = psi4.core.get_global_option("CUBIC_BASIS_TOLERANCE")
        extens     = psi4.core.BasisExtents(basis, epsilon)
        max_points = psi4.core.get_global_option("DFT_BLOCK_MAX_POINTS")        
        npoints    = grid.shape[1]
        nblocks = int(np.floor(npoints/max_points))
        blocks = []

        max_functions = 0
        #Run through full blocks
        idx = 0
        for nb in range(nblocks):
            x = psi4.core.Vector.from_array(grid[0][idx : idx + max_points])
            y = psi4.core.Vector.from_array(grid[1][idx : idx + max_points])
            z = psi4.core.Vector.from_array(grid[2][idx : idx + max_points])
            if if_w:
                w = psi4.core.Vector.from_array(grid[3][idx : idx + max_points])
            else:
                w = psi4.core.Vector.from_array(np.zeros(max_points))  # When w is not necessary and not given

            blocks.append(psi4.core.BlockOPoints(x, y, z, w, extens))
            idx += max_points
            max_functions = max_functions if max_functions > len(blocks[-1].functions_local_to_global()) \
                                          else len(blocks[-1].functions_local_to_global())

        #Run through remaining points
        if idx < npoints:
            x = psi4.core.Vector.from_array(grid[0][idx:])
            y = psi4.core.Vector.from_array(grid[1][idx:])
            z = psi4.core.Vector.from_array(grid[2][idx:])
            if if_w:
                w = psi4.core.Vector.from_array(grid[3][idx:])
            else:
                w = psi4.core.Vector.from_array(np.zeros_like(grid[2][idx:]))  # When w is not necessary and not given
            blocks.append(psi4.core.BlockOPoints(x, y, z, w, extens))

            max_functions = max_functions if max_functions > len(blocks[-1].functions_local_to_global()) \
                                          else len(blocks[-1].functions_local_to_global())

        zero_matrix = psi4.core.Matrix(basis.nbf(), basis.nbf())
        if self.ref == 1:
            point_func = psi4.core.RKSFunctions(basis, max_points, max_functions)
            point_func.set_pointers(zero_matrix)
        else:
            point_func = psi4.core.UKSFunctions(basis, max_points, max_functions)
            point_func.set_pointers(zero_matrix, zero_matrix)

        return blocks, npoints, point_func

    def generate_grids(self, x, y, z):
        """
        Genrates Mesh from 3 separate linear spaces and flatten,
        needed for cubic grid.

        Parameters
        ----------
        grid: tuple of three np.ndarray
            (x, y, z)

        Returns
        -------
        grid: np.ndarray
            shape (3, len(x)*len(y)*len(z)).
        """
        # x,y,z, = grid
        shape = (len(x), len(y), len(z))
        X,Y,Z = np.meshgrid(x, y, z, indexing='ij')
        X = X.reshape((X.shape[0] * X.shape[1] * X.shape[2], 1))
        Y = Y.reshape((Y.shape[0] * Y.shape[1] * Y.shape[2], 1))
        Z = Z.reshape((Z.shape[0] * Z.shape[1] * Z.shape[2], 1))
        grid = np.concatenate((X,Y,Z), axis=1).T

        return grid, shape

    def generate_dft_grid(self, Vpot):
        """
        Extracts DFT spherical grid and weights from wfn object

        Parameters
        ----------
        Vpot: psi4.core.VBase
            Vpot object with dft grid data

        Returns
        -------
        dft_grid: list
            Numpy arrays corresponding to x,y,z, and w.
            Shape: (4, npoints)
        
        """

        nblocks = Vpot.nblocks()
        blocks = [Vpot.get_block(i) for i in range(nblocks)]
        npoints = Vpot.grid().npoints()

        dft_grid = np.zeros((4, npoints))

        offset = 0
        for i_block in blocks:
            b_points = i_block.npoints()
            offset += b_points

            dft_grid[0, offset - b_points : offset] = i_block.x().np
            dft_grid[1, offset - b_points : offset] = i_block.y().np
            dft_grid[2, offset - b_points : offset] = i_block.z().np
            dft_grid[3, offset - b_points : offset] = i_block.w().np

        return dft_grid

    #Quantities on Grid
    def on_grid_ao(self, coeff, grid=None, basis=None, Vpot=None):
        """
        Generates a quantity on the grid given its ao representation.
        *This is the most general function for basis to grid transformation.

        Parameters
        ----------
        coeff: np.ndarray
            Vector/Matrix of quantity on ao basis. Shape: {(num_ao_basis, ), (num_ao_basis, num_ao_basis)}
        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
            grid where density will be computed.
        basis: psi4.core.BasisSet, optional
            The basis set. If not given it will use target wfn.basisset().
        Vpot: psi4.core.VBase
            Vpotential object with info about grid. 
            Provides DFT spherical grid. Only comes to play if no grid is given. 

        Returns
        -------
        coeff_r: np.ndarray Shape: (npoints, )
            Quantity expressed by the coefficient on the given grid 


        """


        if grid is not None:
            if type(grid) is np.ndarray:
                if grid.shape[0] != 3 and grid.shape[0] != 4:
                    raise ValueError("The shape of grid should be (3, npoints) "
                                     "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
                blocks, npoints, points_function = self.grid_to_blocks(grid, basis=basis)
            else:
                blocks, npoints, points_function = grid
        elif grid is None and Vpot is not None:
            nblocks = Vpot.nblocks()
            blocks = [Vpot.get_block(i) for i in range(nblocks)]
            npoints = Vpot.grid().npoints()
            points_function = Vpot.properties()[0]
        else:
            raise ValueError("A grid or a V_potential (DFT grid) must be given.")

        coeff_r = np.zeros((npoints))

        offset = 0
        for i_block in blocks:
            points_function.compute_points(i_block)
            b_points = i_block.npoints()
            offset += b_points
            lpos = np.array(i_block.functions_local_to_global())
            if len(lpos)==0:
                continue
            phi = np.array(points_function.basis_values()["PHI"])[:b_points, :lpos.shape[0]]

            if coeff.ndim == 1:
                l_mat = coeff[(lpos[:])]
                coeff_r[offset - b_points : offset] = contract('pm,m->p', phi, l_mat)
            elif coeff.ndim == 2:
                l_mat = coeff[(lpos[:, None], lpos)]
                coeff_r[offset - b_points : offset] = contract('pm,mn,pn->p', phi, l_mat, phi)

        return coeff_r

    def on_grid_density(self, grid=None,
                              Da=None, 
                              Db=None,
                              Vpot=None):
        """
        Generates Density given grid

        Parameters
        ----------
        Da, Db: np.ndarray
            Alpha, Beta densities. Shape: (num_ao_basis, num_ao_basis)
        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
            grid where density will be computed.
        Vpot: psi4.core.VBase
            Vpotential object with info about grid.
            Provides DFT spherical grid. Only comes to play if no grid is given. 

        Returns
        -------
        density: np.ndarray Shape: (ref, npoints)
            Density on the given grid. 
        """

        if Da is None and Db is None:
            Da = psi4.core.Matrix.from_array(self.Dt[0])
            Db = psi4.core.Matrix.from_array(self.Dt[1])
        else:
            Da = psi4.core.Matrix.from_array(Da)
            Db = psi4.core.Matrix.from_array(Db)

        if self.ref == 2 and Db is None:
            raise ValueError("Db is required for an unrestricted system")

        if grid is not None:
            if type(grid) is np.ndarray:
                if grid.shape[0] != 3 and grid.shape[0] != 4:
                    raise ValueError("The shape of grid should be (3, npoints) "
                                     "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
                blocks, npoints, points_function = self.grid_to_blocks(grid)
            else:
                blocks, npoints, points_function = grid
        elif grid is None and Vpot is not None:
            nblocks = Vpot.nblocks()
            blocks = [Vpot.get_block(i) for i in range(nblocks)]
            npoints = Vpot.grid().npoints()
            points_function = Vpot.properties()[0]
        else:
            raise ValueError("A grid or a V_potential (DFT grid) must be given.")

        
        if self.ref == 1:
            points_function.set_pointers(Da)
            rho_a = points_function.point_values()["RHO_A"]
            density   = np.zeros((npoints))
        if self.ref == 2:
            points_function.set_pointers(Da, Db)
            rho_a = points_function.point_values()["RHO_A"]
            rho_b = points_function.point_values()["RHO_B"]
            density   = np.zeros((npoints, self.ref))

        offset = 0
        for i_block in blocks:
            points_function.compute_points(i_block)
            b_points = i_block.npoints()
            offset += b_points

            if self.ref == 1:
                density[offset - b_points : offset] = rho_a.np[ :b_points]
            else:
                density[offset - b_points : offset, 0] = rho_a.np[ :b_points]
                density[offset - b_points : offset, 1] = rho_b.np[ :b_points]

        return density

    def on_grid_orbitals(self, Ca=None, Cb=None, grid=None, Vpot=None):
        """
        Generates orbitals given grid

        Parameters
        ----------
        Ca, Cb: np.ndarray
            Alpha, Beta Orbital Coefficient Matrix. Shape: (num_ao_basis, num_ao_basis)
        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
            grid where density will be computed
        Vpot: psi4.core.VBase
            Vpotential object with info about grid.
            Provides DFT spherical grid. Only comes to play if no grid is given. 

        Returns
        -------
        orbitals: np.ndarray
            Orbitals on the given grid of size . 
            Shape: (nbasis, npoints, ref)


        """

        if Ca is None and Cb is None:
            Ca = psi4.core.Matrix.from_array(self.Ca)
            Cb = psi4.core.Matrix.from_array(self.Cb)
        else:
            Ca = psi4.core.Matrix.from_array(Ca)
            Cb = psi4.core.Matrix.from_array(Cb)

        if self.ref == 2 and Cb is None:
            raise ValueError("Db is required for an unrestricted system")

        if grid is not None:
            if type(grid) is np.ndarray:
                if grid.shape[0] != 3 and grid.shape[0] != 4:
                    raise ValueError("The shape of grid should be (3, npoints) "
                                     "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
                blocks, npoints, points_function = self.grid_to_blocks(grid)
            else:
                blocks, npoints, points_function = grid
        elif grid is None and Vpot is not None:
            nblocks = Vpot.nblocks()
            blocks = [Vpot.get_block(i) for i in range(nblocks)]
            npoints = Vpot.grid().npoints()
            points_function = Vpot.properties()[0]
        else:
            raise ValueError("A grid or a V_potential (DFT grid) must be given.")

        if self.ref == 1:
            orbitals_r = [np.zeros((npoints)) for i_orb in range(self.nbf)]
            points_function.set_pointers(Ca)
            Ca_np = Ca.np
        if self.ref == 2:
            orbitals_r = [np.zeros((npoints, 2)) for i_orb in range(self.nbf)]
            points_function.set_pointers(Ca, Cb)
            Ca_np = Ca.np
            Cb_np = Cb.np

        offset = 0
        for i_block in blocks:
            points_function.compute_points(i_block)
            b_points = i_block.npoints()
            offset += b_points
            lpos = np.array(i_block.functions_local_to_global())
            if len(lpos)==0:
                continue
            phi = np.array(points_function.basis_values()["PHI"])[:b_points, :lpos.shape[0]]

            for i_orb in range(self.nbf):
                Ca_local = Ca_np[lpos, i_orb]
                if self.ref == 1:
                    orbitals_r[i_orb][offset - b_points : offset] = contract('m, pm -> p', Ca_local, phi)
                else:
                    Cb_local = Cb_np[lpos, i_orb]
                    orbitals_r[i_orb][offset - b_points : offset,0] = contract('m, pm -> p', Ca_local, phi)
                    orbitals_r[i_orb][offset - b_points : offset,1] = contract('m, pm -> p', Cb_local, phi)




        return orbitals_r

    def on_grid_esp(self, Da=None, Db=None, grid=None, Vpot=None, wfn=None):

        """
        Generates EXTERNAL/ESP/HARTREE and Fermi Amaldi Potential on given grid

        Parameters
        ----------
        Da,Db: np.ndarray, opt, shape (nbf, nbf)
            The electron density in the denominator of Hartee potential. If None, the original density matrix
            will be used.
        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
            grid where density will be computed.
        Vpot: psi4.core.VBase
            Vpotential object with info about grid.
            Provides DFT spherical grid. Only comes to play if no grid is given. 
        
        Returns
        -------
        vext, hartree, esp, v_fa: np.ndarray
            External, Hartree, ESP, and Fermi Amaldi potential on the given grid
            Shape: (npoints, )
        """

        if wfn is None:
            wfn = self.wfn

        if Da is not None or Db is not None:
            Da_temp = np.copy(self.wfn.Da().np)
            Db_temp = np.copy(self.wfn.Db().np)
            if Da is not None:
                wfn.Da().np[:] = Da
            if Db is not None:
                wfn.Db().np[:] = Db

        nthreads = psi4.get_num_threads()
        psi4.set_num_threads(1)
        
        if grid is not None:
            if type(grid) is np.ndarray:
                blocks, npoints, points_function = self.grid_to_blocks(grid)
            else:
                blocks, npoints, points_function = grid
        elif grid is None and Vpot is not None:
            nblocks = Vpot.nblocks()
            blocks = [Vpot.get_block(i) for i in range(nblocks)]
            npoints = Vpot.grid().npoints()
        else:
            raise ValueError("A grid or a V_potential (DFT grid) must be given.")

        #Initialize Arrays
        vext = np.zeros(npoints)
        esp  = np.zeros(npoints)

        #Get Atomic Information
        mol_dict = self.mol.to_schema(dtype='psi4')
        natoms = len(mol_dict["elem"])
        indx = [i for i in range(natoms) if self.mol.charge(i) != 0.0]
        natoms = len(indx)
        #Atomic numbers and Atomic positions
        zs = [mol_dict["elez"][i] for i in indx]
        rs = [self.mol.geometry().np[i] for i in indx]

        esp_wfn = psi4.core.ESPPropCalc(wfn)

        #Loop Through blocks
        offset = 0
        with np.errstate(divide='ignore'):
            for i_block in blocks:
                b_points = i_block.npoints()
                offset += b_points
                x = i_block.x().np
                y = i_block.y().np
                z = i_block.z().np

                #EXTERNAL
                for atom in range(natoms):
                    r =  np.sqrt((x-rs[atom][0])**2 + (y-rs[atom][1])**2 + (z-rs[atom][2])**2)
                    vext_temp = - 1.0 * zs[atom] / r
                    vext_temp[np.isinf(vext_temp)] = 0.0
                    vext[offset - b_points : offset] += vext_temp
                #ESP
                xyz = np.concatenate((x[:,None],y[:,None],z[:,None]), axis=1)
                grid_block = psi4.core.Matrix.from_array(xyz)
                esp[offset - b_points : offset] = esp_wfn.compute_esp_over_grid_in_memory(grid_block).np

        #Hartree
        hartree = - 1.0 * (vext + esp)
        v_fa = (1 - 1.0 / (self.nalpha + self.nbeta)) * hartree

        if Da is not None:
            wfn.Da().np[:] = Da_temp
        if Db is not None:
            wfn.Db().np[:] = Db_temp
        psi4.set_num_threads(nthreads)

        return vext, hartree, v_fa, esp

    def on_grid_vxc(self, func_id=1, grid=None, Da=None, Db=None,
                          Vpot=None):
        """
        Generates Vxc given grid

        Parameters
        ----------
        Da, Db: np.ndarray
            Alpha, Beta densities. Shape: (num_ao_basis, num_ao_basis)
        func_id: int
            Functional ID associated with Density Functional Approximationl.
            Full list of functionals: https://www.tddft.org/programs/libxc/functionals/
        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
            grid where density will be computed.
        Vpot: psi4.core.VBase
            Vpotential object with info about grid.
            Provides DFT spherical grid. Only comes to play if no grid is given. 

        Returns
        -------
        VXC: np.ndarray
            Exchange correlation potential on the given grid
            Shape: (npoints, )

        """

        local_functionals = [1,546,549,532,692,641,552,287,307,578,5,24,4,579,308,289,551,
                             22,23,14,11,574,573,554,5900,12,13,25,9,10,27,3,684,683,17,7,
                             28,29,30,31,8,317,2,6,536,537,538,318,577,259,547,548,20,599,43,
                             51,580,50,550
                             ]

        if func_id not in local_functionals:
            raise ValueError("Only LDA fucntionals are supported on the grid")

        if Da is None:
            Da = self.Dt[0]
        if Db is None:
            Db = self.Dt[0]

        if grid is not None:
            if type(grid) is np.ndarray:
                blocks, npoints, points_function = self.grid_to_blocks(grid)
            else:
                blocks, npoints, points_function = grid
            density = self.on_grid_density(Da=Da, Db=Db, grid=grid)
        elif grid is None and Vpot is not None:
            nblocks = Vpot.nblocks()
            blocks = [Vpot.get_block(i) for i in range(nblocks)]
            npoints = Vpot.grid().npoints()
            density = self.on_grid_density(Da=Da, Db=Db, Vpot=Vpot)
        else:
            raise ValueError("A grid or a V_potential (DFT grid) must be given.")

        vxc = np.zeros((npoints, self.ref))
        ingredients = {}
        offset = 0
        for i_block in blocks:
            b_points = i_block.npoints()
            offset += b_points
            if self.ref == 1:
                ingredients["rho"] = density[offset - b_points : offset]
            else:
                ingredients["rho"] = density[offset - b_points : offset, :]

            if self.ref == 1:
                functional = Functional(1, 1)
            else:
                functional = Functional(1, 2) 
            xc_dictionary = functional.compute(ingredients)
            vxc[offset - b_points : offset, :] = xc_dictionary['vrho']

        return np.squeeze(vxc)

    def on_grid_lap_phi(self, 
                       Ca=None, 
                       Cb=None, 
                       grid=None,
                       Vpot=None):
        """
        Generates laplacian of molecular orbitals

        Parameters
        ----------
       Ca, Cb: np.ndarray
            Alpha, Beta Orbital Coefficient Matrix. Shape: (num_ao_basis, num_ao_basis)
        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
            grid where density will be computed.
        Vpot: psi4.core.VBase
            Vpotential object with info about grid.
            Provides DFT spherical grid. Only comes to play if no grid is given. 

        Returns
        -------
        lap_phi: List[np.ndarray]. Where array is of shape (npoints, ref)
            Laplacian of molecular orbitals on the grid
        """

        if Ca is None and Cb is None:
            Ca = psi4.core.Matrix.from_array(self.Ca)
            Cb = psi4.core.Matrix.from_array(self.Cb)
        else:
            Ca = psi4.core.Matrix.from_array(Ca)
            Cb = psi4.core.Matrix.from_array(Cb)

        if self.ref == 2 and Cb is None:
            raise ValueError("Db is required for an unrestricted system")

        if grid is not None:
            if type(grid) is np.ndarray:
                if grid.shape[0] != 3 and grid.shape[0] != 4:
                    raise ValueError("The shape of grid should be (3, npoints) "
                                     "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
                blocks, npoints, points_function = self.grid_to_blocks(grid)
            else:
                blocks, npoints, points_function = grid
        elif grid is None and Vpot is not None:
            nblocks = Vpot.nblocks()
            blocks = [Vpot.get_block(i) for i in range(nblocks)]
            npoints = Vpot.grid().npoints()
            points_function = Vpot.properties()[0]
        else:
            raise ValueError("A grid or a V_potential (DFT grid) must be given.")

        points_function.set_ansatz(2)

        if self.ref == 1:
            points_function.set_pointers(Ca)
            lap_phi = [np.zeros((npoints)) for i_orb in range(self.nbf)]
        else:
            points_function.set_pointers(Ca, Cb)
            lap_phi = [np.zeros((npoints, 2)) for i_orb in range(self.nbf)]

        offset = 0
        for i_block in blocks:
            points_function.compute_points(i_block)
            b_points = i_block.npoints()
            offset += b_points
            lpos = np.array(i_block.functions_local_to_global())
            if len(lpos)==0:
                continue
            
            #Obtain subset of phi_@@ matrices
            lx = np.array(points_function.basis_values()["PHI_XX"])[:b_points, :lpos.shape[0]]
            ly = np.array(points_function.basis_values()["PHI_YY"])[:b_points, :lpos.shape[0]]
            lz = np.array(points_function.basis_values()["PHI_ZZ"])[:b_points, :lpos.shape[0]]

            for i_orb in range(self.nbf):
                Ca_local = Ca.np[lpos, i_orb][:,None]
                
                if self.ref ==1:
                    lap_phi[i_orb][offset - b_points : offset] += ((lx + ly + lz) @ Ca_local)[:,0]
                else:
                    Cb_local = Cb.np[lpos, i_orb][:,None]
                    lap_phi[i_orb][offset - b_points : offset, 0] += ((lx + ly + lz) @ Ca_local)[:,0]
                    lap_phi[i_orb][offset - b_points : offset, 1] += ((lx + ly + lz) @ Cb_local)[:,0]

        return lap_phi

    def on_grid_grad_phi(self, 
                       Ca=None, 
                       Cb=None, 
                       grid=None,
                       Vpot=None):
        """
        Generates laplacian of molecular orbitals

        Parameters
        ----------
       Ca, Cb: np.ndarray
            Alpha, Beta Orbital Coefficient Matrix. Shape: (num_ao_basis, num_ao_basis)
        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
            grid where density will be computed.
        Vpot: psi4.core.VBase
            Vpotential object with info about grid.
            Provides DFT spherical grid. Only comes to play if no grid is given. 

        Returns
        -------
        grad_phi: List[np.ndarray]. Where array is of shape (npoints, ref)
            Gradient of molecular orbitals on the grid
        """

        if Ca is None and Cb is None:
            Ca = psi4.core.Matrix.from_array(self.Ca)
            Cb = psi4.core.Matrix.from_array(self.Cb)
        else:
            Ca = psi4.core.Matrix.from_array(Ca)
            Cb = psi4.core.Matrix.from_array(Cb)

        if self.ref == 2 and Cb is None:
            raise ValueError("Db is required for an unrestricted system")

        if grid is not None:
            if type(grid) is np.ndarray:
                if grid.shape[0] != 3 and grid.shape[0] != 4:
                    raise ValueError("The shape of grid should be (3, npoints) "
                                     "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
                blocks, npoints, points_function = self.grid_to_blocks(grid)
            else:
                blocks, npoints, points_function = grid
        elif grid is None and Vpot is not None:
            nblocks = Vpot.nblocks()
            blocks = [Vpot.get_block(i) for i in range(nblocks)]
            npoints = Vpot.grid().npoints()
            points_function = Vpot.properties()[0]
        else:
            raise ValueError("A grid or a V_potential (DFT grid) must be given.")

        points_function.set_ansatz(2)

        if self.ref == 1:
            points_function.set_pointers(Ca)
            grad_phi = [np.zeros((npoints)) for i_orb in range(self.nbf)]
        else:
            points_function.set_pointers(Ca, Cb)
            grad_phi = [np.zeros((npoints, 2)) for i_orb in range(self.nbf)]

        offset = 0
        for i_block in blocks:
            points_function.compute_points(i_block)
            b_points = i_block.npoints()
            offset += b_points
            lpos = np.array(i_block.functions_local_to_global())
            if len(lpos)==0:
                continue
            
            #Obtain subset of phi_@ matrix
            gx = np.array(points_function.basis_values()["PHI_X"])[:b_points, :lpos.shape[0]]
            gy = np.array(points_function.basis_values()["PHI_Y"])[:b_points, :lpos.shape[0]]
            gz = np.array(points_function.basis_values()["PHI_Z"])[:b_points, :lpos.shape[0]]

            for i_orb in range(self.nbf):
                Ca_local = Ca.np[lpos, i_orb][:,None]
                if self.ref == 1:
                    grad_phi[i_orb][offset - b_points : offset] += ((gx + gy + gz) @ Ca_local)[:,0]
                if self.ref == 2:
                    Cb_local = Cb.np[lpos, i_orb][:,None]
                    grad_phi[i_orb][offset - b_points : offset, 0] += ((gx + gy + gz) @ Ca_local)[:,0]
                    grad_phi[i_orb][offset - b_points : offset, 1] += ((gx + gy + gz) @ Cb_local)[:,0]

        return grad_phi

    def dft_grid_to_fock(self, value, Vpot):
        """For value on DFT spherical grid, Fock matrix is returned.
        VFock_ij = \int dx \phi_i(x) \phi_j(x) value(x)
        
        Parameters:
        -----------
        value: np.ndarray of shape (npoint, ).

        Vpot:psi4.core.VBase
            Vpotential object with info about grid.
            Provides DFT spherical grid. Only comes to play if no grid is given.
        
        Returns:
        ---------
        VFock: np.ndarray of shape (nbasis, nbasis)
        """

        VFock = np.zeros((self.nbf, self.nbf))
        points_func = Vpot.properties()[0]

        i = 0
        # Loop over the blocks
        for b in range(Vpot.nblocks()):
            # Obtain block information
            block = Vpot.get_block(b)
            points_func.compute_points(block)
            npoints = block.npoints()
            lpos = np.array(block.functions_local_to_global())
            if len(lpos) == 0:
                i += npoints
                continue
            # Obtain the grid weight
            w = np.array(block.w())

            # Compute phi!
            phi = np.array(points_func.basis_values()["PHI"])[:npoints, :lpos.shape[0]]

            Vtmp = np.einsum('pb,p,p,pa->ab', phi, value[i:i+npoints], w, phi, optimize=True)

            # Add the temporary back to the larger array by indexing, ensure it is symmetric
            VFock[(lpos[:, None], lpos)] += 0.5 * (Vtmp + Vtmp.T)

            i += npoints
        assert i == value.shape[0], "Did not run through all the points. %i %i" %(i, value.shape[0])
        return VFock

    #Miscellaneous
    def get_basis_set_correction(self, grid):
        return basis_set_correction(self, grid)    



"""
Basis Set Artifact Correction. 
"""

import numpy as np
from warnings import warn
import psi4

def invert_kohn_sham_equations(self, wfn, grid):
    """
    Given a set of canonical kohn-sham orbitals and energies,
    construct vxc by inverting the Kohn-Sham Equations.
    Removal of Basis-Set Artifacts in Kohn–Sham Potentials Recovered from Electron Densities
    Gaiduk + Ryabinkin + Staroverov
    https://pubs.acs.org/doi/abs/10.1021/ct4004146

    Parameters:
    -----------
    wfn: psi4.core.Wavefunction
        Wavefunction from target calculation. 
        Must be obtained from same basis set if future correction will be applied
    grid: np.ndarray. Shape: (3xnpoints)
        Grid used to compute correction. 
        If None, dft grid is used. 

    Returns:
    --------
    vxc_inverted: np.ndarray
        Vxc obtained from inverting kohn sham equations. 
        Equation (3)
    vxc_lda: np.ndarray
        LDA potential from forward calculation. 
    osc_profile: np.ndarray
        Oscillatory profile that corrects inverted potentials. 
        Equation (5)
    """

    if wfn.basisset().name() != self.basis_str:
        warn("""Basis set from calculation is different from Inverter object. \n
                Addition of correction won't fix basis set artifacts.""")
                                
    #Build components on grid:
    if grid is not None:
        #Use user-defined grid
        orb       = self.on_grid_orbitals(Ca=wfn.Ca().np, Cb=wfn.Cb().np, grid=grid)
        lap       = self.on_grid_lap_phi(Ca=wfn.Ca().np, Cb=wfn.Cb().np, grid=grid)
        vex, vha  = self.on_grid_esp(Da=wfn.Da().np, Db=wfn.Db().np, grid=grid)[:2]
        den       = self.on_grid_density(Da=wfn.Da().np, Db=wfn.Db().np, grid=grid)
        eig_a = wfn.epsilon_a_subset("AO", "ALL").np
        eig_b = wfn.epsilon_b_subset("AO", "ALL").np
        
    else:
        #Use DFT grid
        Vpot = wfn.V_potential()
        orb       = self.on_grid_orbitals(Ca=wfn.Ca().np, Cb=wfn.Cb().np, Vpot=Vpot)
        lap       = self.on_grid_lap_phi(Ca=wfn.Ca().np, Cb=wfn.Cb().np, Vpot=Vpot)
        vex, vha  = self.on_grid_esp(Da=wfn.Da().np, Db=wfn.Db().np, Vpot=Vpot)[:2]
        den       = self.on_grid_density(Da=wfn.Da().np, Db=wfn.Db().np, Vpot=Vpot)
        eig_a = wfn.epsilon_a_subset("AO", "ALL").np

    #Build Reversed LDA from orbitals and density
    kinetic = np.zeros_like(den)
    propios = np.zeros_like(den) 

    #Build v_eff. equation (2) for restricted and unrestricted
    if self.ref == 1:
        for i in range(wfn.nalpha()):
            #Alpha orbitals
            kinetic += 2.0 * (0.5 * orb[i] * lap[i])
            propios += 2.0 * eig_a[i] * np.abs(orb[i])**2 
        
        with np.errstate(divide='ignore', invalid='ignore'):
            veff = (kinetic + propios) / den
        vxc_inverted = veff - vha - vex
        
    elif self.ref== 2:
        for i in range(wfn.nalpha()):
            #Alpha orbitals
            kinetic[:,0] += (0.5 * orb[i][:,0] * lap[i][:,0])
            propios[:,0] += eig_a[i] * np.abs(orb[i][:,0])**2 
        for i in range(wfn.nbeta()):
            #Beta orbitals
            kinetic[:,1] += (0.5 * orb[i][:,1] * lap[i][:,1])
            propios[:,1] += eig_b[i] * np.abs(orb[i][:,1])**2 

        with np.errstate(divide='ignore', invalid='ignore'):
            veff = (kinetic + propios) / den
        vxc_inverted = veff - np.repeat(vha[:,None], 2, axis=1) - np.repeat(vex[:,None], 2, axis=1)

    return vxc_inverted

def basis_set_correction(self, grid):
    """
    Obtains basis set correction for inverted potentials as in:
    Removal of Basis-Set Artifacts in Kohn–Sham Potentials Recovered from Electron Densities
    Gaiduk + Ryabinkin + Staroverov
    https://pubs.acs.org/doi/abs/10.1021/ct4004146

    Parameters:
    -----------
    grid: np.ndarray. Shape: (3xnpoints)
        Grid used to compute correction. 
        If None, dft grid is used. 

    Returns:
    --------
    vxc_inverted: np.ndarray
        Vxc obtained from inverting kohn sham equations. 
        Equation (3)
    vxc_lda: np.ndarray
        LDA potential from forward calculation. 
    osc_profile: np.ndarray
        Oscillatory profile that corrects inverted potentials. 
        Equation (5)
    """
    
    #Make a calculation with LDA.
    _, correction_wfn = psi4.energy("svwn/"+self.basis_str, return_wfn=True, molecule=self.mol)
    
    #Build components on grid:
    if grid is not None:
        #Use user-defined grid
        orb       = self.on_grid_orbitals(Ca=correction_wfn.Ca().np, Cb=correction_wfn.Cb().np, grid=grid)
        lap       = self.on_grid_lap_phi(Ca=correction_wfn.Ca().np, Cb=correction_wfn.Cb().np, grid=grid)
        vex, vha  = self.on_grid_esp(Da=correction_wfn.Da().np, Db=correction_wfn.Db().np, grid=grid)[:2]
        vxc_lda   = self.on_grid_vxc(Da=correction_wfn.Da().np, Db=correction_wfn.Db().np, grid=grid)
        den       = self.on_grid_density(Da=correction_wfn.Da().np, Db=correction_wfn.Db().np, grid=grid)
        eig_a = correction_wfn.epsilon_a_subset("AO", "ALL").np
        eig_b = correction_wfn.epsilon_b_subset("AO", "ALL").np
        
    else:
        #Use DFT grid from LDA calculation
        Vpot = correction_wfn.V_potential()
        orb       = self.on_grid_orbitals(Ca=correction_wfn.Ca().np, Cb=correction_wfn.Cb().np, Vpot=Vpot)
        lap       = self.on_grid_lap_phi(Ca=correction_wfn.Ca().np, Cb=correction_wfn.Cb().np, Vpot=Vpot)
        vex, vha  = self.on_grid_esp(Da=correction_wfn.Da().np, Db=correction_wfn.Db().np, Vpot=Vpot)[:2]
        vxc_lda   = self.on_grid_vxc(Da=correction_wfn.Da().np, Db=correction_wfn.Db().np, Vpot=Vpot)
        den       = self.on_grid_density(Da=correction_wfn.Da().np, Db=correction_wfn.Db().np, Vpot=Vpot)
        eig_a = correction_wfn.epsilon_a_subset("AO", "ALL").np
    
    #Build Reversed LDA from orbitals and density
    kinetic = np.zeros_like(den)
    propios = np.zeros_like(den) 

    #Build v_eff. equation (2) for restricted and unrestricted
    if self.ref == 1:
        for i in range(correction_wfn.nalpha()):
            #Alpha orbitals
            kinetic += 2.0 * (0.5 * orb[i] * lap[i])
            propios += 2.0 * eig_a[i] * np.abs(orb[i])**2 
        
        with np.errstate(divide='ignore', invalid='ignore'):
            veff = (kinetic + propios) / den
        vxc_inverted = veff - vha - vex
        osc_profile = vxc_inverted - vxc_lda
        
    elif self.ref== 2:
        for i in range(correction_wfn.nalpha()):
            #Alpha orbitals
            kinetic[:,0] += (0.5 * orb[i][:,0] * lap[i][:,0])
            propios[:,0] += eig_a[i] * np.abs(orb[i][:,0])**2 
        for i in range(correction_wfn.nbeta()):
            #Beta orbitals
            kinetic[:,1] += (0.5 * orb[i][:,1] * lap[i][:,1])
            propios[:,1] += eig_b[i] * np.abs(orb[i][:,1])**2 

        with np.errstate(divide='ignore', invalid='ignore'):
            veff = (kinetic + propios) / den
        vxc_inverted = veff - np.repeat(vha[:,None], 2, axis=1) - np.repeat(vex[:,None], 2, axis=1)
        osc_profile = vxc_inverted - vxc_lda

    return osc_profile


"""
direct.py
"""

import numpy as np
np.seterr(divide='ignore', invalid='ignore')

class Direct():
    def direct_inversion(self, grid, correction=True):
        """
        Given a set of canonical kohn-sham orbitals and energies,
        constructs vxc by inverting the Kohn-Sham Equations.
        If correction is added, it will be performed according to:
        Removal of Basis-Set Artifacts in Kohn–Sham Potentials Recovered from Electron Densities
        Gaiduk + Ryabinkin + Staroverov
        https://pubs.acs.org/doi/abs/10.1021/ct4004146

        Parameters:
        -----------

        grid: np.ndarray. Shape: (3xnpoints)
            Grid used to compute correction. 
            If None, dft grid is used. 
        correction: bool
            Adds correction for spurious basis set artifacts

        Returns:
        --------
        vxc_inverted: np.ndarray
            Vxc obtained from inverting kohn sham equations. 
            Equation (3)
        vxc_lda: np.ndarray
            LDA potential from forward calculation. 
        osc_profile: np.ndarray
            Oscillatory profile that corrects inverted potentials. 
            Equation (5)
        """

        wfn = self.wfn

        #Sanity check. 
        try:
            functional = wfn.functional()
        except Exception as ex:
            raise ValueError(f"{ex}. Direct inversion only available for DFT methods. ")
        if functional.name() == 'HF':
            raise ValueError(f"Direct inversion only available for DFT methods. ")
                                    
        #Build components on grid:
        if grid is not None:
            #Use user-defined grid
            orb       = self.on_grid_orbitals(Ca=wfn.Ca().np, Cb=wfn.Cb().np, grid=grid)
            lap       = self.on_grid_lap_phi(Ca=wfn.Ca().np, Cb=wfn.Cb().np, grid=grid)
            vex, vha  = self.on_grid_esp(Da=wfn.Da().np, Db=wfn.Db().np, grid=grid)[:2]
            den       = self.on_grid_density(Da=wfn.Da().np, Db=wfn.Db().np, grid=grid)
            eig_a = wfn.epsilon_a_subset("AO", "ALL").np
            eig_b = wfn.epsilon_b_subset("AO", "ALL").np
            
            #Calculate correction
            if correction is True:
                osc_profile = self.get_basis_set_correction(grid)

            
        else:
            #Use DFT grid
            Vpot = wfn.V_potential()
            orb       = self.on_grid_orbitals(Ca=wfn.Ca().np, Cb=wfn.Cb().np, Vpot=Vpot)
            lap       = self.on_grid_lap_phi(Ca=wfn.Ca().np, Cb=wfn.Cb().np, Vpot=Vpot)
            vex, vha  = self.on_grid_esp(Da=wfn.Da().np, Db=wfn.Db().np, Vpot=Vpot)[:2]
            den       = self.on_grid_density(Da=wfn.Da().np, Db=wfn.Db().np, Vpot=Vpot)
            eig_a = wfn.epsilon_a_subset("AO", "ALL").np

            #Calculate correction
            if correction is True:
                dft_grid = self.generate_dft_grid(Vpot)
                osc_profile = self.get_basis_set_correction(dft_grid)

        #Build Reversed LDA from orbitals and density
        kinetic = np.zeros_like(den)
        propios = np.zeros_like(den) 

        #Build v_eff. equation (2) for restricted and unrestricted
        if self.ref == 1:
            for i in range(wfn.nalpha()):
                #Alpha orbitals
                kinetic += 2.0 * (0.5 * orb[i] * lap[i])
                propios += 2.0 * eig_a[i] * np.abs(orb[i])**2 
            
            with np.errstate(divide='ignore', invalid='ignore'):
                veff = (kinetic + propios) / den
            vxc_inverted = veff - vha - vex
            
        elif self.ref== 2:
            for i in range(wfn.nalpha()):
                #Alpha orbitals
                kinetic[:,0] += (0.5 * orb[i][:,0] * lap[i][:,0])
                propios[:,0] += eig_a[i] * np.abs(orb[i][:,0])**2 
            for i in range(wfn.nbeta()):
                #Beta orbitals
                kinetic[:,1] += (0.5 * orb[i][:,1] * lap[i][:,1])
                propios[:,1] += eig_b[i] * np.abs(orb[i][:,1])**2 

            with np.errstate(divide='ignore', invalid='ignore'):
                veff = (kinetic + propios) / den
            vxc_inverted = veff - np.repeat(vha[:,None], 2, axis=1) - np.repeat(vex[:,None], 2, axis=1)

        #Add correction
        if correction is True:
            vxc_inverted -= osc_profile

        self.grid.vxc = vxc_inverted
        return vxc_inverted


"""
Unit and regression test for the n2v package.
"""

# Import package, test suite, and other packages as needed
import n2v
import pytest
import sys

def test_n2v_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "n2v" in sys.modules



"""
oucarter.py

Functions associated with Ou-Carter inversion
"""

import numpy as np
from opt_einsum import contract
try:
    import psi4
except:
    pass

class OC():
    """
    Ou-Carter density to potential inversion [1].
    [1] [J. Chem. Theory Comput. 2018, 14, 5680−5689]
    """

    def _get_l_kinetic_energy_density_directly(self, D, C, grid_info=None):
        """
        Calculate $\frac{\tau_L^{KS}}{\rho^{KS}}-\frac{\tau_P^{KS}}{\rho^{KS}}$:
        laplace_rho_temp: $\frac{\nabla^2 \rho}{4}$;
        tauW_temp: $\frac{|\napla \rho|^2}{8|\rho|}$;
        tauLmP_rho: $\frac{|\napla \rho|^2}{8|\rho|^2} - \frac{\nabla^2 \rho}{4\rho}$.

        (i.e. the 2dn and 3rd term in eqn. (17) in [1] over $\rho$.):
        """

        if grid_info is None:
            tauLmP_rho = np.zeros(self.Vpot.grid().npoints())
            nblocks = self.Vpot.nblocks()

            points_func = self.Vpot.properties()[0]
            blocks = None

        else:
            blocks, npoints, points_func = grid_info
            tauLmP_rho = np.zeros(npoints)
            nblocks = len(blocks)

        points_func.set_deriv(2)

        iw = 0
        for l_block in range(nblocks):
            # Obtain general grid information
            if blocks is None:
                l_grid = self.Vpot.get_block(l_block)
            else:
                l_grid = blocks[l_block]
            l_npoints = l_grid.npoints()

            points_func.compute_points(l_grid)
            l_lpos = np.array(l_grid.functions_local_to_global())
            if len(l_lpos) == 0:
                iw += l_npoints
                continue
            l_phi = np.array(points_func.basis_values()["PHI"])[:l_npoints, :l_lpos.shape[0]]
            l_phi_x = np.array(points_func.basis_values()["PHI_X"])[:l_npoints, :l_lpos.shape[0]]
            l_phi_y = np.array(points_func.basis_values()["PHI_Y"])[:l_npoints, :l_lpos.shape[0]]
            l_phi_z = np.array(points_func.basis_values()["PHI_Z"])[:l_npoints, :l_lpos.shape[0]]
            l_phi_xx = np.array(points_func.basis_values()["PHI_XX"])[:l_npoints, :l_lpos.shape[0]]
            l_phi_yy = np.array(points_func.basis_values()["PHI_YY"])[:l_npoints, :l_lpos.shape[0]]
            l_phi_zz = np.array(points_func.basis_values()["PHI_ZZ"])[:l_npoints, :l_lpos.shape[0]]

            lD = D[(l_lpos[:, None], l_lpos)]
            # lC = C[l_lpos, :]

            rho = contract('pm,mn,pn->p', l_phi, lD, l_phi)
            rho_inv = 1/rho

            # Calculate the second term
            laplace_rho_temp = contract('ab,pa,pb->p', lD, l_phi, l_phi_xx + l_phi_yy + l_phi_zz)
            # laplace_rho_temp += contract('pm, mn, pn->p', l_phi_x,lD, l_phi_x)
            # laplace_rho_temp += contract('pm, mn, pn->p', l_phi_y,lD, l_phi_y)
            # laplace_rho_temp += contract('pm, mn, pn->p', l_phi_z,lD, l_phi_z)
            laplace_rho_temp += np.sum((l_phi_x @ lD) * l_phi_x, axis=1)
            laplace_rho_temp += np.sum((l_phi_y @ lD) * l_phi_y, axis=1)
            laplace_rho_temp += np.sum((l_phi_z @ lD) * l_phi_z, axis=1)

            laplace_rho_temp *= 0.25 * 2

            # Calculate the third term
            tauW_temp = contract('pm, mn, pn->p', l_phi, lD, l_phi_x) ** 2
            tauW_temp += contract('pm, mn, pn->p', l_phi, lD, l_phi_y) ** 2
            tauW_temp += contract('pm, mn, pn->p', l_phi, lD, l_phi_z) ** 2
            tauW_temp *= rho_inv * 0.125 * 4

            tauLmP_rho[iw: iw + l_npoints] = (-laplace_rho_temp + tauW_temp) * rho_inv
            iw += l_npoints
        assert iw == tauLmP_rho.shape[0], "Somehow the whole space is not fully integrated."

        return tauLmP_rho

    def _get_optimized_external_potential(self, grid_info, average_alpha_beta=False):
        """
        $
        v^{~}{ext}(r) = \epsilon^{-LDA}(r)
        - \frac{\tau^{LDA}{L}}{n^{LDA}(r)}
        - v_{H}^{LDA}(r) - v_{xc}^{LDA}(r)
        $
        (22) in [1].
        """

        Nalpha = self.nalpha
        Nbeta = self.nbeta

        # SVWN calculation
        wfn_LDA = psi4.energy("SVWN/" + self.basis_str, molecule=self.mol, return_wfn=True)[1]
        Da_LDA = wfn_LDA.Da().np
        Db_LDA = wfn_LDA.Db().np
        Ca_LDA = wfn_LDA.Ca().np
        Cb_LDA = wfn_LDA.Cb().np
        epsilon_a_LDA = wfn_LDA.epsilon_a().np
        epsilon_b_LDA = wfn_LDA.epsilon_b().np
        self.Vpot = wfn_LDA.V_potential()

        vxc_LDA_DFT = self.on_grid_vxc(func_id=1, Da=Da_LDA, Db=Db_LDA, Vpot=self.Vpot)
        vxc_LDA = self.on_grid_vxc(func_id=1,Da=Da_LDA, Db=Db_LDA, grid=grid_info)
        vxc_LDA_DFT += self.on_grid_vxc(func_id=8, Da=Da_LDA, Db=Db_LDA, Vpot=self.Vpot)
        vxc_LDA += self.on_grid_vxc(func_id=8,Da=Da_LDA, Db=Db_LDA, grid=grid_info)
        if self.ref != 1:
            assert vxc_LDA.shape[-1] == 2
            vxc_LDA_beta = vxc_LDA[:,1]
            vxc_LDA = vxc_LDA[:, 0]
            vxc_LDA_DFT_beta = vxc_LDA_DFT[:, 1]
            vxc_LDA_DFT = vxc_LDA_DFT[:, 0]

        # _average_local_orbital_energy() taken from mrks.py.
        e_bar_DFT = self._average_local_orbital_energy(Da_LDA, Ca_LDA[:,:Nalpha], epsilon_a_LDA[:Nalpha])
        e_bar = self._average_local_orbital_energy(Da_LDA, Ca_LDA[:, :Nalpha], epsilon_a_LDA[:Nalpha], grid_info=grid_info)


        tauLmP_rho_DFT = self._get_l_kinetic_energy_density_directly(Da_LDA, Ca_LDA[:,:Nalpha])
        tauLmP_rho = self._get_l_kinetic_energy_density_directly(Da_LDA, Ca_LDA[:,:Nalpha], grid_info=grid_info)

        tauP_rho_DFT = self._pauli_kinetic_energy_density(Da_LDA, Ca_LDA[:,:Nalpha])
        tauP_rho = self._pauli_kinetic_energy_density(Da_LDA, Ca_LDA[:,:Nalpha], grid_info=grid_info)

        tauL_rho_DFT = tauLmP_rho_DFT + tauP_rho_DFT
        tauL_rho = tauLmP_rho + tauP_rho

        vext_opt_no_H_DFT = e_bar_DFT - tauL_rho_DFT - vxc_LDA_DFT
        vext_opt_no_H = e_bar - tauL_rho - vxc_LDA

        J = self.form_jk(Ca_LDA[:,:Nalpha],  Cb_LDA[:,:Nbeta])[0]
        vext_opt_no_H_DFT_Fock = self.dft_grid_to_fock(vext_opt_no_H_DFT, self.Vpot)
        vext_opt_DFT_Fock = vext_opt_no_H_DFT_Fock - J[0] - J[1]

        # # Does vext_opt need a shift?
        # Fock_LDA = self.T + vext_opt_DFT_Fock + J[0] + J[1] + self.dft_grid_to_fock(vxc_LDA_DFT, self.Vpot)
        # eigvecs_a = self.diagonalize(Fock_LDA, self.nalpha)[-1]
        # shift = eigvecs_a[Nalpha-1] - epsilon_a_LDA[Nalpha-1]
        # vext_opt_DFT_Fock -= shift * self.S2
        # print("LDA shift:", shift, eigvecs_a[Nalpha-1], epsilon_a_LDA[Nalpha-1])

        vH = self.on_grid_esp(grid=grid_info, wfn=wfn_LDA)[1]
        vext_opt = vext_opt_no_H - vH
        # vext_opt -= shift

        if self.ref != 1:
            e_bar_DFT_beta = self._average_local_orbital_energy(Db_LDA, Cb_LDA[:,:Nbeta], epsilon_b_LDA[:Nbeta])
            e_bar_beta = self._average_local_orbital_energy(Db_LDA, Cb_LDA[:, :Nbeta], epsilon_b_LDA[:Nbeta], grid_info=grid_info)


            tauLmP_rho_DFT_beta = self._get_l_kinetic_energy_density_directly(Db_LDA, Cb_LDA[:,:Nbeta])
            tauLmP_rho_beta = self._get_l_kinetic_energy_density_directly(Db_LDA, Cb_LDA[:,:Nalpha], grid_info=grid_info)

            tauP_rho_DFT_beta = self._pauli_kinetic_energy_density(Db_LDA, Cb_LDA[:,:Nbeta])
            tauP_rho_beta = self._pauli_kinetic_energy_density(Db_LDA, Cb_LDA[:,:Nbeta], grid_info=grid_info)

            tauL_rho_DFT_beta = tauLmP_rho_DFT_beta + tauP_rho_DFT_beta
            tauL_rho_beta = tauLmP_rho_beta + tauP_rho_beta

            vext_opt_no_H_DFT_beta = e_bar_DFT_beta - tauL_rho_DFT_beta - vxc_LDA_DFT_beta
            vext_opt_no_H_beta = e_bar_beta - tauL_rho_beta - vxc_LDA_beta

            vext_opt_no_H_DFT_Fock_beta = self.dft_grid_to_fock(vext_opt_no_H_DFT_beta, self.Vpot)
            vext_opt_DFT_Fock_beta = vext_opt_no_H_DFT_Fock_beta - J[0] - J[1]

            vext_opt_beta = vext_opt_no_H_beta - vH

            # vext_opt_DFT_Fock = (vext_opt_DFT_Fock + vext_opt_DFT_Fock_beta) * 0.5
            # vext_opt = (vext_opt + vext_opt_beta) * 0.5

            return (vext_opt_DFT_Fock, vext_opt_DFT_Fock_beta), (vext_opt, vext_opt_beta)
        return vext_opt_DFT_Fock, vext_opt
    
    def oucarter(self, maxiter, vxc_grid, D_tol=1e-7,
             eig_tol=1e-4, frac_old=0.5, init="scan"):
        """
        (23) in [1].
        [1] [J. Chem. Theory Comput. 2018, 14, 5680−5689]

        parameters:
        ----------------------
            maxiter: int
                same as opt_max_iter
            vxc_grid: np.ndarray of shape (3, num_grid_points)
                The final result will be represented on this grid
                default: 1e-4
            D_tol: float, opt
                convergence criteria for density matrices.
                default: 1e-7
            eig_tol: float, opt
                convergence criteria for occupied eigenvalue spectrum.
                default: 1e-4
            frac_old: float, opt
                Linear mixing parameter for current vxc and old vxc.
                If 0, no old vxc is mixed in.
                Should be in [0,1)
                default: 0.5.
            init: string, opt
                Initial guess method.
                default: "SCAN"
                1) If None, input wfn info will be used as initial guess.
                2) If "continue" is given, then it will not initialize
                but use the densities and orbitals stored. Meaningly,
                one can run a quick WY calculation as the initial
                guess. This can also be used to user speficified
                initial guess by setting Da, Coca, eigvec_a.
                3) If it's not continue, it would be expecting a
                method name string that works for psi4. A separate psi4 calculation
                would be performed.

        """
        
        # if self.ref != 1:
        #     raise ValueError("Currently only supports Spin-Restricted "
        #                      "calculations since Spin-Unrestricted CI "
        #                      "is not supported by Psi4.")
        
        grid_info = self.grid_to_blocks(vxc_grid)
        if self.ref == 1:
            grid_info[-1].set_pointers(self.wfn.Da())
        else:
            grid_info[-1].set_pointers(self.wfn.Da(), self.wfn.Db())

        if self.ref == 1:
            vext_opt_Fock, vext_opt = self._get_optimized_external_potential(grid_info)
        else:
            (vext_opt_Fock, vext_opt_Fock_beta), (vext_opt, vext_opt_beta) = self._get_optimized_external_potential(grid_info)

        assert self.Vpot is not None
        
        vH0_Fock = self.va
        
        Nalpha = self.nalpha
        Nbeta = self.nbeta
        # Initialization.
        if init is None:
            self.Da = np.copy(self.Dt[0])
            self.Coca = np.copy(self.ct[0])
            self.eigvecs_a = self.wfn.epsilon_a().np[:Nalpha]

            self.Db = np.copy(self.Dt[1])
            self.Cocb = np.copy(self.ct[1])
            self.eigvecs_b = self.wfn.epsilon_b().np[:Nbeta]
        elif init.lower()=="continue":
            pass
        else:
            wfn_temp = psi4.energy(init+"/" + self.basis_str, 
                                   molecule=self.mol, 
                                   return_wfn=True)[1]
            self.Da = np.array(wfn_temp.Da())
            self.Coca = np.array(wfn_temp.Ca())[:, :Nalpha]
            self.eigvecs_a = np.array(wfn_temp.epsilon_a())
            self.Db = np.array(wfn_temp.Db())
            self.Cocb = np.array(wfn_temp.Cb())[:, :Nbeta]
            self.eigvecs_b = np.array(wfn_temp.epsilon_b())
            del wfn_temp

        # nerror = self.on_grid_density(Da=self.Dt[0]-self.Da, Db=self.Dt[1]-self.Da, Vpot=self.Vpot)
        # w = self.Vpot.get_np_xyzw()[-1]
        # nerror = np.sum(np.abs(nerror.T) * w)
        # print("nerror", nerror)

        vxc_old = 0.0
        vxc_old_beta = 0.0
        Da_old = 0.0
        eig_old = 0.0
        tauLmP_rho = self._get_l_kinetic_energy_density_directly(self.Dt[0], self.ct[0][:, :Nalpha])
        if self.ref != 1:
            tauLmP_rho_beta = self._get_l_kinetic_energy_density_directly(self.Dt[1], self.ct[1][:, :Nbeta])

        for OC_step in range(1, maxiter+1):
            tauP_rho = self._pauli_kinetic_energy_density(self.Da, self.Coca)
            # _average_local_orbital_energy() taken from mrks.py.
            e_bar = self._average_local_orbital_energy(self.Da, self.Coca, self.eigvecs_a[:Nalpha])
            shift = self.eigvecs_a[Nalpha - 1] - self.wfn.epsilon_a().np[Nalpha - 1]
            # vxc + vext_opt + vH0
            vxc_extH = e_bar - tauLmP_rho - tauP_rho - shift

            if self.ref != 1:
                tauP_rho_beta = self._pauli_kinetic_energy_density(self.Db, self.Cocb)
                e_bar_beta = self._average_local_orbital_energy(self.Db, self.Cocb, self.eigvecs_b[:Nbeta])
                shift_beta = self.eigvecs_b[Nbeta - 1] - self.wfn.epsilon_b().np[Nbeta - 1]
                # vxc + vext_opt + vH0
                vxc_extH_beta = e_bar_beta - tauLmP_rho_beta - tauP_rho_beta - shift_beta

            Derror = np.linalg.norm(self.Da - Da_old) / self.nbf ** 2
            eerror = (np.linalg.norm(self.eigvecs_a[:Nalpha] - eig_old) / Nalpha)
            if (Derror < D_tol) and (eerror < eig_tol):
                print("KSDFT stops updating.")
                break

            # linear Mixture
            if OC_step != 1:
                vxc_extH = vxc_extH * (1 - frac_old) + vxc_old * frac_old

                if self.ref != 1:
                    vxc_extH_beta = vxc_extH_beta * (1 - frac_old) + vxc_old_beta * frac_old

            vxc_old = np.copy(vxc_extH)
            if self.ref != 1:
                vxc_old_beta = np.copy(vxc_extH_beta)

            # Save old data.
            Da_old = np.copy(self.Da)
            eig_old = np.copy(self.eigvecs_a[:Nalpha])

            Vxc_extH_Fock = self.dft_grid_to_fock(vxc_extH, self.Vpot)
            Vxc_Fock = Vxc_extH_Fock - vext_opt_Fock - vH0_Fock

            if self.ref != 1:
                Vxc_extH_Fock_beta = self.dft_grid_to_fock(vxc_extH_beta, self.Vpot)
                Vxc_Fock_beta = Vxc_extH_Fock_beta - vext_opt_Fock_beta - vH0_Fock

            if self.ref == 1:
                self._diagonalize_with_potential_vFock(v=Vxc_Fock)
            else:
                self._diagonalize_with_potential_vFock(v=(Vxc_Fock, Vxc_Fock_beta))


            print(f"Iter: {OC_step}, Density Change: {Derror:2.2e}, Eigenvalue Change: {eerror:2.2e}.")
            # nerror = self.on_grid_density(Da=self.Dt[0] - self.Da, Db=self.Dt[1] - self.Da, Vpot=self.Vpot)
            # nerror = np.sum(np.abs(nerror.T) * w)
            # print("nerror", nerror)

        # Calculate vxc on grid

        vH0 = self.on_grid_esp(grid=grid_info)[1]
        tauLmP_rho = self._get_l_kinetic_energy_density_directly(self.Dt[0], self.ct[0][:, :Nalpha], grid_info=grid_info)
        tauP_rho = self._pauli_kinetic_energy_density(self.Da, self.Coca, grid_info=grid_info)
        shift = self.eigvecs_a[Nalpha - 1] - self.wfn.epsilon_a().np[Nalpha - 1]
        e_bar = self._average_local_orbital_energy(self.Da, self.Coca, self.eigvecs_a[:Nalpha], grid_info=grid_info)
        if self.ref != 1:
            tauLmP_rho_beta = self._get_l_kinetic_energy_density_directly(self.Dt[1], self.ct[1][:, :Nbeta],
                                                                     grid_info=grid_info)
            tauP_rho_beta = self._pauli_kinetic_energy_density(self.Db, self.Cocb, grid_info=grid_info)
            shift_beta = self.eigvecs_b[Nbeta - 1] - self.wfn.epsilon_b().np[Nbeta - 1]
            e_bar_beta = self._average_local_orbital_energy(self.Db, self.Cocb, self.eigvecs_b[:Nbeta], grid_info=grid_info)

        if self.ref == 1:
            self.grid.vxc = e_bar - tauLmP_rho - tauP_rho - vext_opt - vH0 - shift
            return self.grid.vxc, e_bar, tauLmP_rho, tauP_rho, vext_opt, vH0, shift
        else:
            self.grid.vxc = np.array((e_bar - tauLmP_rho - tauP_rho - vext_opt - vH0 - shift,
                                      e_bar_beta - tauLmP_rho_beta - tauP_rho_beta - vext_opt_beta - vH0 - shift_beta
                                      ))
            return self.grid.vxc, (e_bar, e_bar_beta), (tauLmP_rho, tauLmP_rho_beta), \
                   (tauP_rho,tauP_rho_beta), (vext_opt, vext_opt_beta), vH0, (shift, shift_beta)


import n2v
import numpy as np
import psi4
import pytest

psi4.set_options({"save_jk" : True})
psi4.set_memory(int(2.50e9))

psi4.set_options({
                  'DFT_SPHERICAL_POINTS': 50, 
                  'DFT_RADIAL_POINTS': 50, 
                 })

@pytest.fixture
def wfn():
    Ne = psi4.geometry( 
    """ 
    0 1
    Ne 0.0 0.0 0.0
    noreorient
    nocom
    units bohr
    symmetry c1
    """ )
    psi4.set_options({"reference" : "rhf"}) 
    wfn = psi4.properties("pbe/6-311G",  return_wfn=True, molecule=Ne)[1]

    return wfn

def test_sucess_inversion(wfn):

    x = np.linspace(0.1,1,5)
    y = [0]
    z = [0]

    ine = n2v.Inverter(wfn)
    grid, shape = ine.generate_grids(x,y,z)
    vxc_inverted = ine.invert("direct", grid=grid, correction=False)
    correct_vxc = np.array([-6.70647038, -2.01804532, -1.36686845, -1.04725066, -0.84145449])

    vxc_inverted_corrected = ine.invert("direct", grid=grid, correction=True)
    correct_vxc_inverted = np.array([-4.52326477, -1.52151704, -1.22883751, -0.98372241, -0.75157181])

    assert np.isclose(vxc_inverted.all(), correct_vxc.all())
    assert np.isclose(vxc_inverted_corrected.all(), correct_vxc_inverted.all())

